""" Test for Profiler.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
from os.path import dirname, join
from subprocess import Popen

import pytest
from flaky import flaky

import DIRAC
from DIRAC.Core.Utilities.Profiler import Profiler

# Mark this entire module as slow
pytestmark = pytest.mark.slow


def test_base():
    p = Profiler()
    res = p.pid()
    assert res["OK"] is False
    res = p.status()
    assert res["OK"] is False

    mainProcess = Popen(
        [
            "python",
            join(dirname(DIRAC.__file__), "tests/Utilities/ProcessesCreator_withChildren.py"),
        ]
    )
    time.sleep(1)
    p = Profiler(mainProcess.pid)
    res = p.pid()
    assert res["OK"] is True
    res = p.status()
    assert res["OK"] is True
    res = p.runningTime()
    assert res["OK"] is True
    assert res["Value"] > 0

    res = p.memoryUsage()
    assert res["OK"] is True
    assert res["Value"] > 0
    resWC = p.memoryUsage(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] > 0
    assert resWC["Value"] >= res["Value"]

    res = p.vSizeUsage()
    assert res["OK"] is True
    assert res["Value"] > 0
    resWC = p.vSizeUsage(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] > 0
    assert resWC["Value"] >= res["Value"]

    res = p.vSizeUsage()
    assert res["OK"] is True
    assert res["Value"] > 0
    resWC = p.vSizeUsage(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] > 0
    assert resWC["Value"] >= res["Value"]

    res = p.numThreads()
    assert res["OK"] is True
    assert res["Value"] > 0
    resWC = p.numThreads(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] > 0
    assert resWC["Value"] >= res["Value"]

    res = p.cpuPercentage()
    assert res["OK"] is True
    assert res["Value"] >= 0
    resWC = p.cpuPercentage(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] >= 0
    assert resWC["Value"] >= res["Value"]


@flaky(max_runs=10, min_passes=2)
def test_cpuUsage():
    mainProcess = Popen(
        [
            "python",
            join(dirname(DIRAC.__file__), "tests/Utilities/ProcessesCreator_withChildren.py"),
        ]
    )
    time.sleep(2)
    p = Profiler(mainProcess.pid)
    res = p.pid()
    assert res["OK"] is True
    res = p.status()
    assert res["OK"] is True

    # user
    res = p.cpuUsageUser()
    assert res["OK"] is True
    assert res["Value"] > 0
    resC = p.cpuUsageUser(withChildren=True)
    assert resC["OK"] is True
    assert resC["Value"] > 0
    assert resC["Value"] >= res["Value"]

    res = p.cpuUsageUser()
    assert res["OK"] is True
    assert res["Value"] > 0
    resC = p.cpuUsageUser(withChildren=True)
    assert resC["OK"] is True
    assert resC["Value"] > 0
    assert resC["Value"] >= res["Value"]

    resT = p.cpuUsageUser(withTerminatedChildren=True)
    assert resT["OK"] is True
    assert resT["Value"] > 0
    assert resT["Value"] >= res["Value"]

    resTC = p.cpuUsageUser(withChildren=True, withTerminatedChildren=True)
    assert resTC["OK"] is True
    assert resTC["Value"] > 0
    assert resTC["Value"] >= res["Value"]

    # system
    res = p.cpuUsageSystem()
    assert res["OK"] is True
    assert res["Value"] >= 0
    resWC = p.cpuUsageSystem(withChildren=True)
    assert resWC["OK"] is True
    assert resWC["Value"] >= 0
    assert resWC["Value"] >= res["Value"]

    res = p.cpuUsageSystem()
    assert res["OK"] is True
    assert res["Value"] > 0
    resC = p.cpuUsageSystem(withChildren=True)
    assert resC["OK"] is True
    assert resC["Value"] > 0
    assert resC["Value"] >= res["Value"]

    resT = p.cpuUsageSystem(withTerminatedChildren=True)
    assert resT["OK"] is True
    assert resT["Value"] > 0
    assert resT["Value"] >= res["Value"]

    resTC = p.cpuUsageSystem(withChildren=True, withTerminatedChildren=True)
    assert resTC["OK"] is True
    assert resTC["Value"] > 0
    assert resTC["Value"] >= res["Value"]

    # After this the main process will no-longer exist
    mainProcess.wait()

    res = p.cpuUsageUser()
    assert res["OK"] is False
    assert res["Errno"] == 3



""" Test SLURM
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import pytest
import os
from six.moves import reload_module

from DIRAC import S_OK, gLogger
from DIRAC.Resources.Computing.BatchSystems.SLURM import SLURM


gLogger.setLevel("DEBUG")

executableContent = """
#!/bin/bash

echo "hello world"
"""

expectedContent = """#!/bin/bash
cat > srunExec_1.sh << EOFEXEC

#!/bin/bash

echo "hello world"

EOFEXEC
chmod 755 srunExec_1.sh
srun -l -k srunExec_1.sh
"""

srunOutput = """
1: line1
1: line2
2: line1
1: line3
3: line1
2: line2
3: line2
2: line3
3: line3
"""

srunExpected1 = """
# On node 1

 line1
 line2
 line3
"""


srunExpected2 = """
# On node 3

 line1
 line2
 line3
"""


srunExpected3 = """
# On node 2

 line1
 line2
 line3
"""


srunExpected = [srunExpected1, srunExpected2, srunExpected3]


normalOutput = """
line1
line2
line3
"""


normalExpected = """
line1
line2
line3
"""


@pytest.mark.parametrize(
    "expectedContent",
    [
        (expectedContent),
    ],
)
def test_generateWrapper(mocker, expectedContent):
    """Test generateWrapper()"""
    mocker.patch("DIRAC.Resources.Computing.BatchSystems.SLURM.random.randrange", return_value=1)
    slurm = SLURM()

    executableFile = "executableFile.sh"
    with open(executableFile, "w") as f:
        f.write(executableContent)

    res = slurm._generateSrunWrapper(executableFile)
    # Make sure a wrapper file has been generated and is executable
    assert res == expectedContent

    os.remove(executableFile)


@pytest.mark.parametrize(
    "numberOfNodes, outputContent, expectedContent",
    [
        ("3-5", srunOutput, srunExpected),
        ("1", normalOutput, normalExpected),
    ],
)
def test_getJobOutputFiles(numberOfNodes, outputContent, expectedContent):
    """Test getJobOutputFiles()"""
    slurm = SLURM()

    # We remove the '\n' at the beginning/end of the file because there are not present in reality
    outputContent = outputContent.strip()
    # We only remove the '\n' at the beginning because processOutput adds a '\n' at the end
    expectedContent = [i.lstrip() for i in expectedContent]

    outputFile = "./1234.out"
    with open(outputFile, "w") as f:
        f.write(outputContent)

    errorFile = "./1234.err"
    with open(errorFile, "w") as f:
        f.write(outputContent)

    batchDict = {
        "JobIDList": ["1234"],
        "OutputDir": ".",
        "ErrorDir": ".",
        "NumberOfNodes": numberOfNodes,
    }
    result = slurm.getJobOutputFiles(**batchDict)
    assert result["Status"] == 0

    output = result["Jobs"]["1234"]["Output"]
    error = result["Jobs"]["1234"]["Error"]
    assert output == outputFile
    assert error == errorFile

    with open(outputFile, "r") as f:
        wrapperContent = f.read()
    for srunLines in expectedContent:
        assert srunLines in wrapperContent

    os.remove(outputFile)
    os.remove(errorFile)



"""
  This modules defines two functions that can be used in place of the ``RPCClient`` and
  ``TransferClient`` to transparently switch to ``https``.

  Example::

    from DIRAC.Core.Tornado.Client.ClientSelector import RPCClientSelector as RPCClient
    myService = RPCClient("Framework/MyService")
    myService.doSomething()
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import functools

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceURL
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.Tornado.Client.TornadoClient import TornadoClient


sLog = gLogger.getSubLogger(__name__)


def ClientSelector(disetClient, *args, **kwargs):  # We use same interface as RPCClient
    """
    Select the correct Client (either RPC or Transfer ), instantiate it, and return it.

    The selection is based on the URL:

    * either it contains the protocol, in which case we make a choice
    * or it is in the form <Component/Service>, in which case we resolve first

    This is a generic function. You should rather use :py:class:`.RPCClientSelector`
    or :py:class:`.TransferClientSelector`


    In principle, the only place for this class to be used is in
    :py:class:`DIRAC.Core.Base.Client.Client`, since it is the only
    one supposed to instantiate an :py:class:`DIRAC.Core.Base.DISET.RPCClient.RPCClient`

    :params disetClient: the DISET class to be instantiated, so either
        :py:class:`DIRAC.Core.Base.DISET.RPCClient.RPCClient`
        or :py:class:`DIRAC.Core.Base.DISET.TransferClient.TransferClient`
    :param args: Whatever ``disetClient`` takes as args, but the first one is
        always the URL we want to rely on.
        It can be either "system/service" or "dips://domain:port/system/service"
    :param kwargs: This can contain:

      * Whatever ``disetClient`` takes.
      * httpsClient: specific class inheriting from TornadoClient

    """

    # We detect if we need to use a specific class for the HTTPS client

    tornadoClient = kwargs.pop("httpsClient", TornadoClient)

    # We have to make URL resolution BEFORE the RPCClient or TornadoClient to determine which one we want to use
    # URL is defined as first argument (called serviceName) in RPCClient

    try:
        serviceName = args[0]
        sLog.debug("Trying to autodetect client for %s" % serviceName)

        # If we are not already given a URL, resolve it
        if serviceName.startswith(("http", "dip")):
            completeUrl = serviceName
        else:
            completeUrl = getServiceURL(serviceName)
            sLog.debug("URL resolved: %s" % completeUrl)

        if completeUrl.startswith("http"):
            sLog.verbose("Using HTTPS for service %s" % serviceName)
            rpc = tornadoClient(*args, **kwargs)
        else:
            rpc = disetClient(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        # If anything went wrong in the resolution, we return default RPCClient
        # So the behaviour is exactly the same as before implementation of Tornado
        sLog.warn("Could not select DISET or Tornado client", "%s" % repr(e))
        rpc = disetClient(*args, **kwargs)
    return rpc


# Client to use for RPC selection
RPCClientSelector = functools.partial(ClientSelector, RPCClient)

# Client to use for Transfer selection
TransferClientSelector = functools.partial(ClientSelector, TransferClient)



#!/usr/bin/env python
"""
Script that synchronizes the resources described on the CS with the RSS.
By default, it sets their Status to `Unknown`, StatusType to `all` and
reason to `Synchronized`. However, it can copy over the status on the CS to
the RSS. Important: If the StatusType is not defined on the CS, it will set
it to Banned !
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import version, gLogger, exit as DIRACExit, S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

subLogger = None
switchDict = {}
DEFAULT_STATUS = ""


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    Script.registerSwitch(
        "", "init", "Initialize the element to the status in the CS ( applicable for StorageElements )"
    )
    Script.registerSwitch("", "element=", "Element family to be Synchronized ( Site, Resource or Node ) or `all`")
    Script.registerSwitch("", "defaultStatus=", "Default element status if not given in the CS")


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """
    usageMessage = "DIRAC %s\n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if args:
        subLogger.error("Found the following positional args '%s', but we only accept switches" % args)
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    switches = dict(Script.getUnprocessedSwitches())

    # Default values
    switches.setdefault("element", None)
    switches.setdefault("defaultStatus", "Banned")
    if not switches["element"] in ("all", "Site", "Resource", "Node", None):
        subLogger.error("Found %s as element switch" % switches["element"])
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return switches


def synchronize():
    """
    Given the element switch, adds rows to the <element>Status tables with Status
    `Unknown` and Reason `Synchronized`.
    """
    global DEFAULT_STATUS

    from DIRAC.ResourceStatusSystem.Utilities import Synchronizer

    synchronizer = Synchronizer.Synchronizer(defaultStatus=DEFAULT_STATUS)

    if switchDict["element"] in ("Site", "all"):
        subLogger.info("Synchronizing Sites")
        res = synchronizer._syncSites()
        if not res["OK"]:
            return res

    if switchDict["element"] in ("Resource", "all"):
        subLogger.info("Synchronizing Resource")
        res = synchronizer._syncResources()
        if not res["OK"]:
            return res

    if switchDict["element"] in ("Node", "all"):
        subLogger.info("Synchronizing Nodes")
        res = synchronizer._syncNodes()
        if not res["OK"]:
            return res

    return S_OK()


def initSites():
    """
    Initializes Sites statuses taking their values from the "SiteMask" table of "JobDB" database.
    """
    from DIRAC.WorkloadManagementSystem.Client.WMSAdministratorClient import WMSAdministratorClient
    from DIRAC.ResourceStatusSystem.Client import ResourceStatusClient

    rssClient = ResourceStatusClient.ResourceStatusClient()

    sites = WMSAdministratorClient().getAllSiteMaskStatus()

    if not sites["OK"]:
        subLogger.error(sites["Message"])
        DIRACExit(1)

    for site, elements in sites["Value"].items():
        result = rssClient.addOrModifyStatusElement(
            "Site",
            "Status",
            name=site,
            statusType="all",
            status=elements[0],
            elementType=site.split(".")[0],
            tokenOwner="rs_svc",
            reason="dirac-rss-sync",
        )
        if not result["OK"]:
            subLogger.error(result["Message"])
            DIRACExit(1)

    return S_OK()


def initSEs():
    """
    Initializes SEs statuses taking their values from the CS.
    """
    from DIRAC import gConfig
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
    from DIRAC.ResourceStatusSystem.Utilities import CSHelpers, RssConfiguration
    from DIRAC.ResourceStatusSystem.PolicySystem import StateMachine
    from DIRAC.ResourceStatusSystem.Client import ResourceStatusClient

    # WarmUp local copy
    CSHelpers.warmUp()

    subLogger.info("Initializing SEs")

    rssClient = ResourceStatusClient.ResourceStatusClient()

    statuses = StateMachine.RSSMachine(None).getStates()
    statusTypes = RssConfiguration.RssConfiguration().getConfigStatusType("StorageElement")
    reason = "dirac-rss-sync"

    subLogger.debug(statuses)
    subLogger.debug(statusTypes)

    for se in DMSHelpers().getStorageElements():

        subLogger.debug(se)

        opts = gConfig.getOptionsDict("/Resources/StorageElements/%s" % se)
        if not opts["OK"]:
            subLogger.warn(opts["Message"])
            continue
        opts = opts["Value"]

        subLogger.debug(opts)

        # We copy the list into a new object to remove items INSIDE the loop !
        statusTypesList = statusTypes[:]

        for statusType, status in opts.items():

            # Sanity check...
            if statusType not in statusTypesList:
                continue

            # Transforms statuses to RSS terms
            if status in ("NotAllowed", "InActive"):
                status = "Banned"

            if status not in statuses:
                subLogger.error("%s not a valid status for %s - %s" % (status, se, statusType))
                continue

            # We remove from the backtracking
            statusTypesList.remove(statusType)

            subLogger.debug([se, statusType, status, reason])
            result = rssClient.addOrModifyStatusElement(
                "Resource",
                "Status",
                name=se,
                statusType=statusType,
                status=status,
                elementType="StorageElement",
                tokenOwner="rs_svc",
                reason=reason,
            )

            if not result["OK"]:
                subLogger.error("Failed to modify")
                subLogger.error(result["Message"])
                continue

        # Backtracking: statusTypes not present on CS
        for statusType in statusTypesList:

            result = rssClient.addOrModifyStatusElement(
                "Resource",
                "Status",
                name=se,
                statusType=statusType,
                status=DEFAULT_STATUS,
                elementType="StorageElement",
                tokenOwner="rs_svc",
                reason=reason,
            )
            if not result["OK"]:
                subLogger.error("Error in backtracking for %s,%s,%s" % (se, statusType, status))
                subLogger.error(result["Message"])

    return S_OK()


def run():
    """
    Main function of the script
    """

    result = synchronize()
    if not result["OK"]:
        subLogger.error(result["Message"])
        DIRACExit(1)

    if "init" in switchDict:

        if switchDict.get("element") == "Site":
            result = initSites()
            if not result["OK"]:
                subLogger.error(result["Message"])
                DIRACExit(1)

        if switchDict.get("element") == "Resource":
            result = initSEs()
            if not result["OK"]:
                subLogger.error(result["Message"])
                DIRACExit(1)


@Script()
def main():
    global subLogger
    global switchDict
    global DEFAULT_STATUS

    subLogger = gLogger.getSubLogger(__file__)
    registerSwitches()
    registerUsageMessage()
    switchDict = parseSwitches()
    DEFAULT_STATUS = switchDict.get("defaultStatus", "Banned")

    # Run script
    run()

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()



""".. module:: Test_Mail

Test cases for DIRAC.Core.Utilities.DAG module.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.Mail import Mail

__RCSID__ = "$Id $"


def test_createEmail():
    m = Mail()
    res = m._create("address@dirac.org")
    assert not res["OK"]

    m._subject = "subject"
    m._fromAddress = "from@dirac.org"
    m._mailAddress = "address@dirac.org"
    m._message = "This is a message"
    res = m._create("address@dirac.org")
    assert res["OK"]
    assert res["Value"].__dict__["_headers"] == [
        ("Content-Type", "multipart/mixed"),
        ("MIME-Version", "1.0"),
        ("Subject", "subject"),
        ("From", "from@dirac.org"),
        ("To", "address@dirac.org"),
    ]


def test_compareEmails(monkeypatch):
    # The hostname on GitHub actions can change randomly so mock it
    monkeypatch.setattr("socket.getfqdn", lambda: "localhost.example")

    m1 = Mail()
    m2 = Mail()
    assert m1 == m2, (m1.__dict__, m2.__dict__)

    m1 = Mail()
    m1._subject = "subject"
    m1._fromAddress = "from@dirac.org"
    m1._mailAddress = "address@dirac.org"
    m1._message = "This is a message"
    m2 = Mail()
    m2._subject = "subject"
    m2._fromAddress = "from@dirac.org"
    m2._mailAddress = "address@dirac.org"
    m2._message = "This is a message"
    assert m1 == m2
    m3 = Mail()
    m3._subject = "subject"
    m3._fromAddress = "from@dirac.org"
    m3._mailAddress = "address@dirac.org"
    m3._message = "This is a message a bit different"
    assert m1 != m3

    s = {m1, m2}
    assert len(s) == 1
    s.add(m2)
    assert len(s) == 1
    s.add(m3)
    assert len(s) == 2
    s.add(m3)
    assert len(s) == 2



"""
  Defines the plugin to take storage space information given by WLCG Accounting Json
  https://twiki.cern.ch/twiki/bin/view/LCG/AccountingTaskForce#Storage_Space_Accounting
  https://twiki.cern.ch/twiki/pub/LCG/AccountingTaskForce/storage_service_v4.txt
  https://docs.google.com/document/d/1yzCvKpxsbcQC5K9MyvXc-vBF1HGPBk4vhjw3MEXoXf8

  When this is used, the OccupancyLFN has to be the full path on the storage, and not just the LFN
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import json
import os
import tempfile
import shutil
import errno
import gfal2  # pylint: disable=import-error

from DIRAC import S_OK, S_ERROR


class WLCGAccountingJson(object):
    """.. class:: WLCGAccountingJson

    Occupancy plugin to return the space information given by WLCG Accouting Json
    """

    def __init__(self, se):
        self.se = se
        self.log = se.log.getSubLogger("WLCGAccountingJson")
        self.name = self.se.name

    def _downloadJsonFile(self, occupancyLFN, filePath):
        """Download the json file at the location

        :param occupancyLFN: lfn for the file
        :param filePath: destination path for the file

        """
        for storage in self.se.storages:
            try:
                ctx = gfal2.creat_context()
                params = ctx.transfer_parameters()
                params.overwrite = True
                res = storage.updateURL(occupancyLFN)
                if not res["OK"]:
                    continue
                occupancyURL = res["Value"]
                ctx.filecopy(params, occupancyURL, "file://" + filePath)
                return
            except gfal2.GError as e:
                detailMsg = "Failed to copy file %s to destination url %s: [%d] %s" % (
                    occupancyURL,
                    filePath,
                    e.code,
                    e.message,
                )
                self.log.debug("Exception while copying", detailMsg)
                continue

    def getOccupancy(self, **kwargs):
        """Returns the space information given by WLCG Accouting Json

        :returns: S_OK with dict (keys: SpaceReservation, Total, Free)
        """
        occupancyLFN = kwargs["occupancyLFN"]

        if not occupancyLFN:
            return S_ERROR("Failed to get occupancyLFN")

        tmpDirName = tempfile.mkdtemp()
        filePath = os.path.join(tmpDirName, os.path.basename(occupancyLFN))

        self._downloadJsonFile(occupancyLFN, filePath)

        if not os.path.isfile(filePath):
            return S_ERROR("No WLCGAccountingJson file of %s is downloaded." % (self.name))

        with open(filePath, "r") as path:
            occupancyDict = json.load(path)

        # delete temp dir
        shutil.rmtree(tmpDirName)

        try:
            storageShares = occupancyDict["storageservice"]["storageshares"]
        except KeyError as e:
            return S_ERROR(
                errno.ENOMSG, "Issue finding storage shares. %s in %s at %s." % (repr(e), occupancyLFN, self.name)
            )

        spaceReservation = self.se.options.get("SpaceReservation")

        # get storageshares in WLCGAccountingJson file
        storageSharesSR = None
        if spaceReservation:
            for storageshare in storageShares:
                if storageshare.get("name") == spaceReservation:
                    storageSharesSR = storageshare
                    break
        else:
            self.log.debug(
                "Could not find SpaceReservation in CS, and get storageShares and spaceReservation from WLCGAccoutingJson."
            )
            shareLen = []
            for storage in self.se.storages:
                basePath = storage.getParameters()["Path"]
                for share in storageShares:
                    shareLen.append((share, len(os.path.commonprefix([share["path"][0], basePath]))))
            storageSharesSR = max(shareLen, key=lambda x: x[1])[0]
            spaceReservation = storageSharesSR.get("name")

        sTokenDict = {}
        sTokenDict["SpaceReservation"] = spaceReservation
        try:
            sTokenDict["Total"] = storageSharesSR["totalsize"]
            sTokenDict["Free"] = storageSharesSR.get("freesize", sTokenDict["Total"] - storageSharesSR["usedsize"])
        except KeyError as e:
            return S_ERROR(
                errno.ENOMSG,
                "Issue finding Total or Free space left. %s in %s storageshares." % (repr(e), spaceReservation),
            )

        return S_OK(sTokenDict)



########################################################################
# File: RegisterOperation.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/19 13:55:14
########################################################################
""" :mod: RegisterFile

    ==================

    .. module: RegisterFile

    :synopsis: register operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    RegisterFile operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file RegisterOperation.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/19 13:55:24
# @brief Definition of RegisterOperation class.

# # imports
from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class RegisterFile(OperationHandlerBase):
    """
    .. class:: RegisterOperation

    RegisterFile operation handler

    :param self: self reference
    :param ~DIRAC.RequestManagementSystem.Client.Operation.Operation operation: Operation instance
    :param str csPath: CS path for this handler

    """

    def __init__(self, operation=None, csPath=None):
        """c'tor"""
        OperationHandlerBase.__init__(self, operation, csPath)

    def __call__(self):
        """call me maybe"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # RegisterFile specific monitor info
            gMonitor.registerActivity(
                "RegisterAtt", "Attempted file registrations", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterOK", "Successful file registrations", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterFail", "Failed file registrations", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # counter for failed files
        failedFiles = 0
        # # catalog(s) to use
        catalogs = self.operation.Catalog
        if catalogs:
            catalogs = [cat.strip() for cat in catalogs.split(",")]
        dm = DataManager(catalogs=catalogs)
        # # get waiting files
        waitingFiles = self.getWaitingFilesList()

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(waitingFiles)))

        # # loop over files
        for opFile in waitingFiles:

            if not self.rmsMonitoring:
                gMonitor.addMark("RegisterAtt", 1)

            # # get LFN
            lfn = opFile.LFN
            # # and others
            fileTuple = (lfn, opFile.PFN, opFile.Size, self.operation.targetSEList[0], opFile.GUID, opFile.Checksum)
            # # call DataManager
            registerFile = dm.registerFile(fileTuple)
            # # check results
            if not registerFile["OK"] or lfn in registerFile["Value"]["Failed"]:

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("RegisterFail", 1)
                # self.dataLoggingClient().addFileRecord(
                #     lfn, "RegisterFail", ','.join(catalogs) if catalogs else "all catalogs", "", "RegisterFile")

                reason = str(
                    registerFile.get("Message", registerFile.get("Value", {}).get("Failed", {}).get(lfn, "Unknown"))
                )
                errorStr = "failed to register LFN"
                opFile.Error = "%s: %s" % (errorStr, reason)
                if "GUID already registered" in reason:
                    opFile.Status = "Failed"
                    self.log.error(errorStr, "%s: %s" % (lfn, reason))
                elif "File already registered with no replicas" in reason:
                    self.log.warn(errorStr, "%s: %s, will remove it and retry" % (lfn, reason))
                    dm.removeFile(lfn)
                else:
                    self.log.warn(errorStr, "%s: %s" % (lfn, reason))
                failedFiles += 1

            else:

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                else:
                    gMonitor.addMark("RegisterOK", 1)
                # self.dataLoggingClient().addFileRecord(
                #     lfn, "Register", ','.join(catalogs) if catalogs else "all catalogs", "", "RegisterFile")

                self.log.verbose(
                    "file %s has been registered at %s" % (lfn, ",".join(catalogs) if catalogs else "all catalogs")
                )
                opFile.Status = "Done"

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        # # final check
        if failedFiles:
            self.log.warn("all files processed, %s files failed to register" % failedFiles)
            self.operation.Error = "some files failed to register"
            return S_ERROR(self.operation.Error)

        return S_OK()



#! /usr/bin/env python
########################################################################
# File :    dirac-admin-kill-pilot
# Author :  A.T.
########################################################################
"""
Kill the specified pilot
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("PilotRef: pilot reference")
    _, args = Script.parseCommandLine(ignoreErrors=True)

    pilotRef = args[0]

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0

    result = diracAdmin.killPilot(pilotRef)
    if not result["OK"]:
        DIRAC.gLogger.error("Failed to kill pilot", pilotRef)
        DIRAC.gLogger.error(result["Message"])
        exitCode = 1

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-configure
# Author :  Ricardo Graciani
########################################################################
"""
  Main script to write dirac.cfg for a new DIRAC installation and initial download of CAs and CRLs
  if necessary.

  To be used by VO specific scripts to configure new DIRAC installations

  Additionally all options can all be passed inside a .cfg file, see the `--cfg` option.
  The following options are recognized::

      Setup
      ConfigurationServer
      IncludeAllServers
      Gateway
      SiteName
      CEName
      VirtualOrganization
      UseServerCertificate
      SkipCAChecks
      SkipCADownload
      UseVersionsDir
      Architecture
      LocalSE
      LogLevel

  Setup and ConfigurationServer(Gateway) is mandatory options.

  As in any other script command line option take precedence over .cfg files passed as arguments.
  The combination of both is written into the installed dirac.cfg.

  Notice: It will not overwrite exiting info in current dirac.cfg if it exists.

  Example:
    $ dirac-configure -d
                      -S LHCb-Development
                      -C 'dips://lhcbprod.pic.es:9135/Configuration/Server'
                      -W 'dips://lhcbprod.pic.es:9135'
                      --SkipCAChecks

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import sys
import os

import six

import DIRAC
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers import cfgInstallPath, cfgPath, Registry
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.FrameworkSystem.Client.BundleDeliveryClient import BundleDeliveryClient

__RCSID__ = "$Id$"


class Params(object):
    def __init__(self):
        self.logLevel = None
        self.setup = None
        self.configurationServer = None
        self.includeAllServers = False
        self.gatewayServer = None
        self.siteName = None
        self.useServerCert = False
        self.skipCAChecks = False
        self.skipCADownload = False
        self.useVersionsDir = False
        self.architecture = None
        self.localSE = None
        self.ceName = None
        self.vo = None
        self.update = False
        self.outputFile = ""
        self.skipVOMSDownload = False
        self.extensions = ""

    def setGateway(self, optionValue):
        self.gatewayServer = optionValue
        self.setServer(self.gatewayServer + "/Configuration/Server")
        DIRAC.gConfig.setOptionValue(cfgInstallPath("Gateway"), self.gatewayServer)
        return DIRAC.S_OK()

    def setOutput(self, optionValue):
        self.outputFile = optionValue
        return DIRAC.S_OK()

    def setServer(self, optionValue):
        self.configurationServer = optionValue
        DIRAC.gConfig.setOptionValue("/DIRAC/Configuration/Servers", self.configurationServer)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("ConfigurationServer"), self.configurationServer)
        return DIRAC.S_OK()

    def setAllServers(self, optionValue):
        self.includeAllServers = True

    def setSetup(self, optionValue):
        self.setup = optionValue
        DIRAC.gConfig.setOptionValue("/DIRAC/Setup", self.setup)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("Setup"), self.setup)
        return DIRAC.S_OK()

    def setSiteName(self, optionValue):
        self.siteName = optionValue
        Script.localCfg.addDefaultEntry("/LocalSite/Site", self.siteName)
        DIRAC.__siteName = False
        DIRAC.gConfig.setOptionValue(cfgInstallPath("SiteName"), self.siteName)
        return DIRAC.S_OK()

    def setCEName(self, optionValue):
        self.ceName = optionValue
        DIRAC.gConfig.setOptionValue(cfgInstallPath("CEName"), self.ceName)
        return DIRAC.S_OK()

    def setServerCert(self, optionValue):
        self.useServerCert = True
        DIRAC.gConfig.setOptionValue(cfgInstallPath("UseServerCertificate"), self.useServerCert)
        return DIRAC.S_OK()

    def setSkipCAChecks(self, optionValue):
        self.skipCAChecks = True
        DIRAC.gConfig.setOptionValue(cfgInstallPath("SkipCAChecks"), self.skipCAChecks)
        return DIRAC.S_OK()

    def setSkipCADownload(self, optionValue):
        self.skipCADownload = True
        DIRAC.gConfig.setOptionValue(cfgInstallPath("SkipCADownload"), self.skipCADownload)
        return DIRAC.S_OK()

    def setSkipVOMSDownload(self, optionValue):
        self.skipVOMSDownload = True
        DIRAC.gConfig.setOptionValue(cfgInstallPath("SkipVOMSDownload"), self.skipVOMSDownload)
        return DIRAC.S_OK()

    def setUseVersionsDir(self, optionValue):
        self.useVersionsDir = True
        DIRAC.gConfig.setOptionValue(cfgInstallPath("UseVersionsDir"), self.useVersionsDir)
        return DIRAC.S_OK()

    def setArchitecture(self, optionValue):
        self.architecture = optionValue
        Script.localCfg.addDefaultEntry("/LocalSite/Architecture", self.architecture)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("Architecture"), self.architecture)
        return DIRAC.S_OK()

    def setLocalSE(self, optionValue):
        self.localSE = optionValue
        Script.localCfg.addDefaultEntry("/LocalSite/LocalSE", self.localSE)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("LocalSE"), self.localSE)
        return DIRAC.S_OK()

    def setVO(self, optionValue):
        self.vo = optionValue
        Script.localCfg.addDefaultEntry("/DIRAC/VirtualOrganization", self.vo)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("VirtualOrganization"), self.vo)
        return DIRAC.S_OK()

    def forceUpdate(self, optionValue):
        self.update = True
        return DIRAC.S_OK()

    def setExtensions(self, optionValue):
        self.extensions = optionValue
        DIRAC.gConfig.setOptionValue("/DIRAC/Extensions", self.extensions)
        DIRAC.gConfig.setOptionValue(cfgInstallPath("Extensions"), self.extensions)
        return DIRAC.S_OK()


def _runConfigurationWizard(setups, defaultSetup):
    """The implementation of the configuration wizard"""
    from prompt_toolkit import prompt, print_formatted_text, HTML
    from prompt_toolkit.completion import FuzzyWordCompleter

    # It makes no sense to have suggestions if there is no default so adjust the message accordingly
    msg = ""
    if setups:
        msg = "press tab for suggestions"
        if defaultSetup:
            msg = "default</b> <green>%s</green><b>, %s" % (defaultSetup, msg)
        msg = " (%s)" % msg
    # Get the Setup
    setup = prompt(
        HTML("<b>Choose a DIRAC Setup%s:</b>\n" % msg),
        completer=FuzzyWordCompleter(list(setups)),
    )
    if defaultSetup and not setup:
        setup = defaultSetup
    if setup not in setups:
        print_formatted_text(HTML("Unknown setup <yellow>%s</yellow> chosen" % setup))
        confirm = prompt(HTML("<b>Are you sure you want to continue?</b> "), default="n")
        if confirm.lower() not in ["y", "yes"]:
            return None

    # Get the URL to the master CS
    csURL = prompt(HTML("<b>Choose a configuration server URL (leave blank for default):</b>\n"))
    if not csURL:
        csURL = setups[setup]

    # Confirm
    print_formatted_text(
        HTML(
            "<b>Configuration is:</b>\n"
            + "  * <b>Setup:</b> <green>%s</green>\n" % setup
            + "  * <b>Configuration server:</b> <green>%s</green>\n" % csURL
        )
    )
    confirm = prompt(HTML("<b>Are you sure you want to continue?</b> "), default="y")
    if confirm.lower() in ["y", "yes"]:
        return setup, csURL
    else:
        return None


def runConfigurationWizard(params):
    """Interactively configure DIRAC using metadata from installed extensions"""
    import subprocess
    from prompt_toolkit import prompt, print_formatted_text, HTML
    from DIRAC.Core.Utilities.Extensions import extensionsByPriority, getExtensionMetadata

    for extension in extensionsByPriority():
        extensionMetadata = getExtensionMetadata(extension)
        if extensionMetadata.get("primary_extension", False):
            break
    defaultSetup = extensionMetadata.get("default_setup", "")
    setups = extensionMetadata.get("setups", {})

    # Run the wizard
    try:
        # Get the user's password and create a proxy so we can download from the CS
        while True:
            userPasswd = prompt(u"Enter Certificate password: ", is_password=True)
            result = subprocess.run(  # pylint: disable=no-member
                ["dirac-proxy-init", "--nocs", "--no-upload", "--pwstdin"],
                input=userPasswd,
                encoding="utf-8",
                check=False,
            )
            if result.returncode == 0:
                break
            print_formatted_text(HTML("<red>Wizard failed, retrying...</red> (press Control + C to exit)\n"))

        print_formatted_text()

        # Ask the user for the appropriate configuration settings
        while True:
            result = _runConfigurationWizard(setups, defaultSetup)
            if result:
                break
            print_formatted_text(HTML("<red>Wizard failed, retrying...</red> (press Control + C to exit)\n"))
    except KeyboardInterrupt:
        print_formatted_text(HTML("<red>Cancelled</red>"))
        sys.exit(1)

    # Apply the arguments to the params object
    setup, csURL = result
    params.setSetup(setup)
    params.setServer(csURL)
    params.setSkipCAChecks(True)

    # Do the actual configuration
    runDiracConfigure(params)

    # Generate a new proxy without passing --nocs
    result = subprocess.run(  # pylint: disable=no-member
        ["dirac-proxy-init", "--pwstdin"],
        input=userPasswd,
        encoding="utf-8",
        check=False,
    )
    sys.exit(result.returncode)


@Script()
def main():
    Script.disableCS()
    params = Params()
    if six.PY3 and len(sys.argv) < 2:
        runConfigurationWizard(params)
    else:
        return runDiracConfigure(params)


def runDiracConfigure(params):
    Script.registerSwitch("S:", "Setup=", "Set <setup> as DIRAC setup", params.setSetup)
    Script.registerSwitch("e:", "Extensions=", "Set <extensions> as DIRAC extensions", params.setExtensions)
    Script.registerSwitch("C:", "ConfigurationServer=", "Set <server> as DIRAC configuration server", params.setServer)
    Script.registerSwitch("I", "IncludeAllServers", "include all Configuration Servers", params.setAllServers)
    Script.registerSwitch("n:", "SiteName=", "Set <sitename> as DIRAC Site Name", params.setSiteName)
    Script.registerSwitch("N:", "CEName=", "Determiner <sitename> from <cename>", params.setCEName)
    Script.registerSwitch("V:", "VO=", "Set the VO name", params.setVO)

    Script.registerSwitch("W:", "gateway=", "Configure <gateway> as DIRAC Gateway for the site", params.setGateway)

    Script.registerSwitch("U", "UseServerCertificate", "Configure to use Server Certificate", params.setServerCert)
    Script.registerSwitch("H", "SkipCAChecks", "Configure to skip check of CAs", params.setSkipCAChecks)
    Script.registerSwitch("D", "SkipCADownload", "Configure to skip download of CAs", params.setSkipCADownload)
    Script.registerSwitch(
        "M", "SkipVOMSDownload", "Configure to skip download of VOMS info", params.setSkipVOMSDownload
    )

    Script.registerSwitch("v", "UseVersionsDir", "Use versions directory", params.setUseVersionsDir)

    Script.registerSwitch("A:", "Architecture=", "Configure /Architecture=<architecture>", params.setArchitecture)
    Script.registerSwitch("L:", "LocalSE=", "Configure LocalSite/LocalSE=<localse>", params.setLocalSE)

    Script.registerSwitch(
        "F",
        "ForceUpdate",
        "Force Update of cfg file (i.e. dirac.cfg) (otherwise nothing happens if dirac.cfg already exists)",
        params.forceUpdate,
    )

    Script.registerSwitch("O:", "output=", "output configuration file", params.setOutput)

    Script.parseCommandLine(ignoreErrors=True)

    if not params.logLevel:
        params.logLevel = DIRAC.gConfig.getValue(cfgInstallPath("LogLevel"), "")
        if params.logLevel:
            DIRAC.gLogger.setLevel(params.logLevel)
    else:
        DIRAC.gConfig.setOptionValue(cfgInstallPath("LogLevel"), params.logLevel)

    if not params.gatewayServer:
        newGatewayServer = DIRAC.gConfig.getValue(cfgInstallPath("Gateway"), "")
        if newGatewayServer:
            params.setGateway(newGatewayServer)

    if not params.configurationServer:
        newConfigurationServer = DIRAC.gConfig.getValue(cfgInstallPath("ConfigurationServer"), "")
        if newConfigurationServer:
            params.setServer(newConfigurationServer)

    if not params.includeAllServers:
        newIncludeAllServer = DIRAC.gConfig.getValue(cfgInstallPath("IncludeAllServers"), False)
        if newIncludeAllServer:
            params.setAllServers(True)

    if not params.setup:
        newSetup = DIRAC.gConfig.getValue(cfgInstallPath("Setup"), "")
        if newSetup:
            params.setSetup(newSetup)

    if not params.siteName:
        newSiteName = DIRAC.gConfig.getValue(cfgInstallPath("SiteName"), "")
        if newSiteName:
            params.setSiteName(newSiteName)

    if not params.ceName:
        newCEName = DIRAC.gConfig.getValue(cfgInstallPath("CEName"), "")
        if newCEName:
            params.setCEName(newCEName)

    if not params.useServerCert:
        newUserServerCert = DIRAC.gConfig.getValue(cfgInstallPath("UseServerCertificate"), False)
        if newUserServerCert:
            params.setServerCert(newUserServerCert)

    if not params.skipCAChecks:
        newSkipCAChecks = DIRAC.gConfig.getValue(cfgInstallPath("SkipCAChecks"), False)
        if newSkipCAChecks:
            params.setSkipCAChecks(newSkipCAChecks)

    if not params.skipCADownload:
        newSkipCADownload = DIRAC.gConfig.getValue(cfgInstallPath("SkipCADownload"), False)
        if newSkipCADownload:
            params.setSkipCADownload(newSkipCADownload)

    if not params.useVersionsDir:
        newUseVersionsDir = DIRAC.gConfig.getValue(cfgInstallPath("UseVersionsDir"), False)
        if newUseVersionsDir:
            params.setUseVersionsDir(newUseVersionsDir)
            # Set proper Defaults in configuration (even if they will be properly overwrite by gComponentInstaller
            instancePath = os.path.dirname(os.path.dirname(DIRAC.rootPath))
            rootPath = os.path.join(instancePath, "pro")
            DIRAC.gConfig.setOptionValue(cfgInstallPath("InstancePath"), instancePath)
            DIRAC.gConfig.setOptionValue(cfgInstallPath("RootPath"), rootPath)

    if not params.architecture:
        newArchitecture = DIRAC.gConfig.getValue(cfgInstallPath("Architecture"), "")
        if newArchitecture:
            params.setArchitecture(newArchitecture)

    if not params.vo:
        newVO = DIRAC.gConfig.getValue(cfgInstallPath("VirtualOrganization"), "")
        if newVO:
            params.setVO(newVO)

    if not params.extensions:
        newExtensions = DIRAC.gConfig.getValue(cfgInstallPath("Extensions"), "")
        if newExtensions:
            params.setExtensions(newExtensions)

    DIRAC.gLogger.notice("Executing: %s " % (" ".join(sys.argv)))
    DIRAC.gLogger.notice('Checking DIRAC installation at "%s"' % DIRAC.rootPath)

    if params.update:
        if params.outputFile:
            DIRAC.gLogger.notice("Will update the output file %s" % params.outputFile)
        else:
            DIRAC.gLogger.notice("Will update %s" % DIRAC.gConfig.diracConfigFilePath)

    if params.setup:
        DIRAC.gLogger.verbose("/DIRAC/Setup =", params.setup)
    if params.vo:
        DIRAC.gLogger.verbose("/DIRAC/VirtualOrganization =", params.vo)
    if params.configurationServer:
        DIRAC.gLogger.verbose("/DIRAC/Configuration/Servers =", params.configurationServer)

    if params.siteName:
        DIRAC.gLogger.verbose("/LocalSite/Site =", params.siteName)
    if params.architecture:
        DIRAC.gLogger.verbose("/LocalSite/Architecture =", params.architecture)
    if params.localSE:
        DIRAC.gLogger.verbose("/LocalSite/localSE =", params.localSE)

    if not params.useServerCert:
        DIRAC.gLogger.verbose("/DIRAC/Security/UseServerCertificate =", "no")
        # Being sure it was not there before
        Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
        Script.localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "no")
    else:
        DIRAC.gLogger.verbose("/DIRAC/Security/UseServerCertificate =", "yes")
        # Being sure it was not there before
        Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
        Script.localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")

    host = DIRAC.gConfig.getValue(cfgInstallPath("Host"), "")
    if host:
        DIRAC.gConfig.setOptionValue(cfgPath("DIRAC", "Hostname"), host)

    if params.skipCAChecks:
        DIRAC.gLogger.verbose("/DIRAC/Security/SkipCAChecks =", "yes")
        # Being sure it was not there before
        Script.localCfg.deleteOption("/DIRAC/Security/SkipCAChecks")
        Script.localCfg.addDefaultEntry("/DIRAC/Security/SkipCAChecks", "yes")
    else:
        # Necessary to allow initial download of CA's
        if not params.skipCADownload:
            DIRAC.gConfig.setOptionValue("/DIRAC/Security/SkipCAChecks", "yes")
    if not params.skipCADownload:
        Script.enableCS()
        try:
            dirName = os.path.join(DIRAC.rootPath, "etc", "grid-security", "certificates")
            mkDir(dirName)
        except Exception:
            DIRAC.gLogger.exception()
            DIRAC.gLogger.fatal("Fail to create directory:", dirName)
            DIRAC.exit(-1)
        try:
            bdc = BundleDeliveryClient()
            result = bdc.syncCAs()
            if result["OK"]:
                result = bdc.syncCRLs()
        except Exception as e:
            DIRAC.gLogger.error("Failed to sync CAs and CRLs: %s" % str(e))

        Script.localCfg.deleteOption("/DIRAC/Security/SkipCAChecks")

    if params.ceName or params.siteName:
        # This is used in the pilot context, we should have a proxy, or a certificate, and access to CS
        if params.useServerCert:
            # Being sure it was not there before
            Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
            Script.localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
        Script.enableCS()
        # Get the site resource section
        gridSections = DIRAC.gConfig.getSections("/Resources/Sites/")
        if not gridSections["OK"]:
            DIRAC.gLogger.warn("Could not get grid sections list")
            grids = []
        else:
            grids = gridSections["Value"]
        # try to get siteName from ceName or Local SE from siteName using Remote Configuration
        for grid in grids:
            siteSections = DIRAC.gConfig.getSections("/Resources/Sites/%s/" % grid)
            if not siteSections["OK"]:
                DIRAC.gLogger.warn("Could not get %s site list" % grid)
                sites = []
            else:
                sites = siteSections["Value"]

            if not params.siteName:
                if params.ceName:
                    for site in sites:
                        res = DIRAC.gConfig.getSections("/Resources/Sites/%s/%s/CEs/" % (grid, site), [])
                        if not res["OK"]:
                            DIRAC.gLogger.warn("Could not get %s CEs list" % site)
                        if params.ceName in res["Value"]:
                            params.siteName = site
                            break
            if params.siteName:
                DIRAC.gLogger.notice("Setting /LocalSite/Site = %s" % params.siteName)
                Script.localCfg.addDefaultEntry("/LocalSite/Site", params.siteName)
                DIRAC.__siteName = False
                if params.ceName:
                    DIRAC.gLogger.notice("Setting /LocalSite/GridCE = %s" % params.ceName)
                    Script.localCfg.addDefaultEntry("/LocalSite/GridCE", params.ceName)

                if not params.localSE and params.siteName in sites:
                    params.localSE = getSEsForSite(params.siteName)
                    if params.localSE["OK"] and params.localSE["Value"]:
                        params.localSE = ",".join(params.localSE["Value"])
                        DIRAC.gLogger.notice("Setting /LocalSite/LocalSE =", params.localSE)
                        Script.localCfg.addDefaultEntry("/LocalSite/LocalSE", params.localSE)
                    break

    if params.gatewayServer:
        DIRAC.gLogger.verbose("/DIRAC/Gateways/%s =" % DIRAC.siteName(), params.gatewayServer)
        Script.localCfg.addDefaultEntry("/DIRAC/Gateways/%s" % DIRAC.siteName(), params.gatewayServer)

    # Create the local cfg if it is not yet there
    if not params.outputFile:
        params.outputFile = DIRAC.gConfig.diracConfigFilePath
    params.outputFile = os.path.abspath(params.outputFile)
    if not os.path.exists(params.outputFile):
        configDir = os.path.dirname(params.outputFile)
        mkDir(configDir)
        params.update = True
        DIRAC.gConfig.dumpLocalCFGToFile(params.outputFile)

    if params.includeAllServers:
        # We need user proxy or server certificate to continue in order to get all the CS URLs
        if not params.useServerCert:
            Script.enableCS()
            result = getProxyInfo()
            if not result["OK"]:
                DIRAC.gLogger.notice("Configuration is not completed because no user proxy is available")
                DIRAC.gLogger.notice("Create one using dirac-proxy-init and execute again with -F option")
                return 1
        else:
            Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
            # When using Server Certs CA's will be checked, the flag only disables initial download
            # this will be replaced by the use of SkipCADownload
            Script.localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
            Script.enableCS()

        DIRAC.gConfig.setOptionValue("/DIRAC/Configuration/Servers", ",".join(DIRAC.gConfig.getServersList()))
        DIRAC.gLogger.verbose("/DIRAC/Configuration/Servers =", ",".join(DIRAC.gConfig.getServersList()))

    if params.useServerCert:
        # always removing before dumping
        Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
        Script.localCfg.deleteOption("/DIRAC/Security/SkipCAChecks")
        Script.localCfg.deleteOption("/DIRAC/Security/SkipVOMSDownload")

    if params.update:
        DIRAC.gConfig.dumpLocalCFGToFile(params.outputFile)

    # ## LAST PART: do the vomsdir/vomses magic

    # This has to be done for all VOs in the installation

    if params.skipVOMSDownload:
        return 0

    result = Registry.getVOMSServerInfo()
    if not result["OK"]:
        return 1

    error = ""
    vomsDict = result["Value"]
    for vo in vomsDict:
        voName = vomsDict[vo]["VOMSName"]
        vomsDirPath = os.path.join(DIRAC.rootPath, "etc", "grid-security", "vomsdir", voName)
        vomsesDirPath = os.path.join(DIRAC.rootPath, "etc", "grid-security", "vomses")
        for path in (vomsDirPath, vomsesDirPath):
            mkDir(path)
        vomsesLines = []
        for vomsHost in vomsDict[vo].get("Servers", {}):
            hostFilePath = os.path.join(vomsDirPath, "%s.lsc" % vomsHost)
            try:
                DN = vomsDict[vo]["Servers"][vomsHost]["DN"]
                CA = vomsDict[vo]["Servers"][vomsHost]["CA"]
                port = vomsDict[vo]["Servers"][vomsHost]["Port"]
                if not DN or not CA or not port:
                    DIRAC.gLogger.error("DN = %s" % DN)
                    DIRAC.gLogger.error("CA = %s" % CA)
                    DIRAC.gLogger.error("Port = %s" % port)
                    DIRAC.gLogger.error("Missing Parameter for %s" % vomsHost)
                    continue
                with open(hostFilePath, "wt") as fd:
                    fd.write("%s\n%s\n" % (DN, CA))
                vomsesLines.append('"%s" "%s" "%s" "%s" "%s" "24"' % (voName, vomsHost, port, DN, voName))
                DIRAC.gLogger.notice("Created vomsdir file %s" % hostFilePath)
            except Exception:
                DIRAC.gLogger.exception("Could not generate vomsdir file for host", vomsHost)
                error = "Could not generate vomsdir file for VO %s, host %s" % (voName, vomsHost)
        try:
            vomsesFilePath = os.path.join(vomsesDirPath, voName)
            with open(vomsesFilePath, "wt") as fd:
                fd.write("%s\n" % "\n".join(vomsesLines))
            DIRAC.gLogger.notice("Created vomses file %s" % vomsesFilePath)
        except Exception:
            DIRAC.gLogger.exception("Could not generate vomses file")
            error = "Could not generate vomses file for VO %s" % voName

    if params.useServerCert:
        Script.localCfg.deleteOption("/DIRAC/Security/UseServerCertificate")
        # When using Server Certs CA's will be checked, the flag only disables initial download
        # this will be replaced by the use of SkipCADownload
        Script.localCfg.deleteOption("/DIRAC/Security/SkipCAChecks")

    if error:
        return 1

    return 0


if __name__ == "__main__":
    exitCode = main()
    Script.localCfg.deleteOption("/DIRAC/Security/SkipCAChecks")
    sys.exit(exitCode)



"""Component Supervision Agent for monitoring agent, executor or service behaviour and intervene if
necessary.

This agent is designed to supervise the Agents, Executors and Services, and restarts them in case
they get stuck. It can only control components running on the same machine as the agent. One agent
per server is needed.

* The agent checks the age of the log file and if it is deemed too old will kill the agent so that
  it is restarted automatically. (Option RestartAgents)
* Executors will only be restarted if there are jobs in checking status (Option RestartExecutors)
* Services will be restarted if they don't answer a ping RPC (Option RestartServices)

* Check for running and stopped components and ensure they have the proper status as defined in the
  CS Registry/Hosts/_HOST_/[Running|Stopped] sections. (Option ControlComponents)
* If desired also service URLs can automatically be added or removed from the Configuration (Option CommitURLs)

The configuration for Running and Stopped components are two sub-sections in Registry/Hosts/<Host>::

    Running
    {
      Configuration__Server =
      Framework__SystemAdministrator =
      Framework__ComponentSupervisionAgent =
    }
    Stopped
    {
      DataManagement__FileCatalog2 =
      Framework__Monitoring =
    }

By moving from one to the other section we can make the ComponentSupervisionAgent Stop/Start the given
component. Values for the entries in the list are ignored, Syntax is ``<System>__<ComponentName>``.

For full functioning of the Agent a few additional permissions have to be granted for the ``Operator`` role.
In the ``SystemAdministrator/Authorization`` section of the relevant setup::

  getOverallStatus = Operator
  stopComponent = Operator
  startComponent = Operator

``getOverallStatus`` is needed for basic functioning of the Agent. ``stopComponent`` or ``startComponent`` are only
needed if ``ControlComponents`` is enabled.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ComponentSupervisionAgent
  :end-before: ##END
  :dedent: 2
  :caption: ComponentSupervisionAgent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
from collections import defaultdict
from datetime import datetime
from functools import partial

import os
import socket
import psutil

# from DIRAC
from DIRAC import S_OK, S_ERROR, gConfig, rootPath
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.ConfigurationSystem.Client.Helpers import Path
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Base.Client import Client
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.FrameworkSystem.Client.SystemAdministratorClient import SystemAdministratorClient
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

__RCSID__ = "$Id$"
AGENT_NAME = "Framework/ComponentSupervisionAgent"

# Define units
HOUR = 3600
MINUTES = 60
SECONDS = 1

# Define constant
NO_CHECKING_JOBS = "NO_CHECKING_JOBS"
CHECKING_JOBS = "CHECKING_JOBS"
NO_RESTART = "NO_RESTART"


class ComponentSupervisionAgent(AgentModule):
    """ComponentSupervisionAgent class."""

    def __init__(self, *args, **kwargs):
        """Initialize the agent, clients, default values."""
        AgentModule.__init__(self, *args, **kwargs)
        self.name = "ComponentSupervisionAgent"
        self.setup = "DIRAC-Production"
        self.enabled = False
        self.restartAgents = False
        self.restartExecutors = False
        self.restartServices = False
        self.controlComponents = False
        self.commitURLs = False
        self.doNotRestartInstancePattern = ["RequestExecutingAgent"]
        self.diracLocation = rootPath

        self.sysAdminClient = SystemAdministratorClient(socket.getfqdn())
        self.jobMonClient = JobMonitoringClient()
        self.nClient = NotificationClient()
        self.csAPI = None
        self.agents = dict()
        self.executors = dict()
        self.services = dict()
        self._tornadoPort = "8443"
        self.errors = list()
        self.accounting = defaultdict(dict)

        self.addressTo = []
        self.addressFrom = ""
        self.emailSubject = "ComponentSupervisionAgent on %s" % socket.getfqdn()

    def logError(self, errStr, varMsg=""):
        """Append errors to a list, which is sent in email notification."""
        self.log.error(errStr, varMsg)
        self.errors.append(errStr + " " + varMsg)

    def beginExecution(self):
        """Reload the configurations before every cycle."""
        self.setup = self.am_getOption("Setup", self.setup)
        self.enabled = self.am_getOption("EnableFlag", self.enabled)
        self.restartAgents = self.am_getOption("RestartAgents", self.restartAgents)
        self.restartExecutors = self.am_getOption("RestartExecutors", self.restartExecutors)
        self.restartServices = self.am_getOption("RestartServices", self.restartServices)
        self.addressTo = self.am_getOption("MailTo", self.addressTo)
        self.addressFrom = self.am_getOption("MailFrom", self.addressFrom)
        self.controlComponents = self.am_getOption("ControlComponents", self.controlComponents)
        self.commitURLs = self.am_getOption("CommitURLs", self.commitURLs)
        self.doNotRestartInstancePattern = self.am_getOption(
            "DoNotRestartInstancePattern", self.doNotRestartInstancePattern
        )

        self.csAPI = CSAPI()

        res = self.getRunningInstances(instanceType="Agents")
        if not res["OK"]:
            return S_ERROR("Failure to get running agents")
        self.agents = res["Value"]

        res = self.getRunningInstances(instanceType="Executors")
        if not res["OK"]:
            return S_ERROR("Failure to get running executors")
        self.executors = res["Value"]

        res = self.getRunningInstances(instanceType="Services")
        if not res["OK"]:
            return S_ERROR("Failure to get running services")
        self.services = res["Value"]

        self.accounting.clear()
        return S_OK()

    def sendNotification(self):
        """Send email notification about changes done in the last cycle."""
        if not (self.errors or self.accounting):
            return S_OK()

        emailBody = ""
        rows = []
        for instanceName, val in self.accounting.items():
            rows.append(
                [[instanceName], [val.get("Treatment", "No Treatment")], [str(val.get("LogAge", "Not Relevant"))]]
            )

        if rows:
            columns = ["Instance", "Treatment", "Log File Age (Minutes)"]
            emailBody += printTable(columns, rows, printOut=False, numbering=False, columnSeparator=" | ")

        if self.errors:
            emailBody += "\n\nErrors:"
            emailBody += "\n".join(self.errors)

        self.log.notice("Sending Email:\n" + emailBody)
        for address in self.addressTo:
            res = self.nClient.sendMail(address, self.emailSubject, emailBody, self.addressFrom, localAttempt=False)
            if not res["OK"]:
                self.log.error("Failure to send Email notification to ", address)
                continue

        self.errors = []
        self.accounting.clear()

        return S_OK()

    def getRunningInstances(self, instanceType="Agents", runitStatus="Run"):
        """Return a dict of running agents, executors or services.

        Key is component's name, value contains dict with PollingTime, PID, Port, Module, RunitStatus, LogFileLocation

        :param str instanceType: 'Agents', 'Executors', 'Services'
        :param str runitStatus: Return only those instances with given RunitStatus or 'All'
        :returns: Dictionary of running instances
        """
        res = self.sysAdminClient.getOverallStatus()
        if not res["OK"]:
            self.logError("Failure to get %s from system administrator client" % instanceType, res["Message"])
            return res

        val = res["Value"][instanceType]
        runningComponents = defaultdict(dict)
        for system, components in val.items():
            for componentName, componentInfo in components.items():
                if componentInfo["Setup"] and componentInfo["Installed"]:
                    if runitStatus != "All" and componentInfo["RunitStatus"] != runitStatus:
                        continue
                    for option, default in (("PollingTime", HOUR), ("Port", None), ("Protocol", None)):
                        runningComponents[componentName][option] = self._getComponentOption(
                            instanceType, system, componentName, option, default
                        )
                        # remove empty values so we can use defaults in _getURL
                        if not runningComponents[componentName][option]:
                            runningComponents[componentName].pop(option)
                    runningComponents[componentName]["LogFileLocation"] = os.path.join(
                        self.diracLocation, "runit", system, componentName, "log", "current"
                    )
                    runningComponents[componentName]["PID"] = componentInfo["PID"]
                    runningComponents[componentName]["Module"] = componentInfo["Module"]
                    runningComponents[componentName]["RunitStatus"] = componentInfo["RunitStatus"]
                    runningComponents[componentName]["System"] = system

        return S_OK(runningComponents)

    def _getComponentOption(self, instanceType, system, componentName, option, default):
        """Get component option from DIRAC CS, using components' base classes methods."""
        componentPath = PathFinder.getComponentSection(
            system=system,
            component=componentName,
            setup=self.setup,
            componentCategory=instanceType,
        )
        if instanceType != "Agents":
            return gConfig.getValue(Path.cfgPath(componentPath, option), default)
        # deal with agent configuration
        componentLoadModule = gConfig.getValue(Path.cfgPath(componentPath, "Module"), componentName)
        fullComponentName = Path.cfgPath(system, componentName)
        fullComponentLoadName = Path.cfgPath(system, componentLoadModule)
        return AgentModule(fullComponentName, fullComponentLoadName).am_getOption(option, default)

    def on_terminate(self, componentName, process):
        """Execute callback when a process terminates gracefully."""
        self.log.info("%s's process with ID: %s has been terminated successfully" % (componentName, process.pid))

    def execute(self):
        """Execute checks for agents, executors, services."""
        for instanceType in ("executor", "agent", "service"):
            for name, options in getattr(self, instanceType + "s").items():
                # call checkAgent, checkExecutor, checkService
                res = getattr(self, "check" + instanceType.capitalize())(name, options)
                if not res["OK"]:
                    self.logError("Failure when checking %s" % instanceType, "%s, %s" % (name, res["Message"]))

        res = self.componentControl()
        if not res["OK"]:
            if "Stopped does not exist" not in res["Message"] and "Running does not exist" not in res["Message"]:
                self.logError("Failure to control components", res["Message"])

        if not self.errors:
            res = self.checkURLs()
            if not res["OK"]:
                self.logError("Failure to check URLs", res["Message"])
        else:
            self.logError("Something was wrong before, not checking URLs this time")

        self.sendNotification()

        if self.errors:
            return S_ERROR("Error during this cycle, check log")

        return S_OK()

    @staticmethod
    def getLastAccessTime(logFileLocation):
        """Return the age of log file."""
        lastAccessTime = 0
        try:
            lastAccessTime = os.path.getmtime(logFileLocation)
            lastAccessTime = datetime.fromtimestamp(lastAccessTime)
        except OSError as e:
            return S_ERROR("Failed to access logfile %s: %r" % (logFileLocation, e))

        now = datetime.now()
        age = now - lastAccessTime
        return S_OK(age)

    def restartInstance(self, pid, instanceName, enabled):
        """Kill a process which is then restarted automatically."""
        if not (self.enabled and enabled):
            self.log.info("Restarting is disabled, please restart %s manually" % instanceName)
            self.accounting[instanceName]["Treatment"] = "Please restart it manually"
            return S_OK(NO_RESTART)

        if any(pattern in instanceName for pattern in self.doNotRestartInstancePattern):
            self.log.info("Restarting for %s is disabled, please restart it manually" % instanceName)
            self.accounting[instanceName]["Treatment"] = "Please restart it manually"
            return S_OK(NO_RESTART)

        try:
            componentProc = psutil.Process(int(pid))
            processesToTerminate = componentProc.children(recursive=True)
            processesToTerminate.append(componentProc)

            for proc in processesToTerminate:
                proc.terminate()

            _gone, alive = psutil.wait_procs(
                processesToTerminate, timeout=5, callback=partial(self.on_terminate, instanceName)
            )
            for proc in alive:
                self.log.info("Forcefully killing process %s" % proc.pid)
                proc.kill()

            return S_OK()

        except psutil.Error as err:
            self.logError("Exception occurred in terminating processes", "%s" % err)
            return S_ERROR()

    def checkService(self, serviceName, options):
        """Ping the service, restart if the ping does not respond."""
        url = self._getURL(serviceName, options)
        self.log.info("Pinging service", url)
        pingRes = Client().ping(url=url)
        if not pingRes["OK"]:
            self.log.info("Failure pinging service: %s: %s" % (url, pingRes["Message"]))
            res = self.restartInstance(int(options["PID"]), serviceName, self.restartServices)
            if not res["OK"]:
                return res
            if res["Value"] != NO_RESTART:
                self.accounting[serviceName]["Treatment"] = "Successfully Restarted"
                self.log.info("Service %s has been successfully restarted" % serviceName)
        self.log.info("Service responded OK")
        return S_OK()

    def checkAgent(self, agentName, options):
        """Check the age of agent's log file, if it is too old then restart the agent."""
        pollingTime, currentLogLocation, pid = (options["PollingTime"], options["LogFileLocation"], options["PID"])
        self.log.info("Checking Agent: %s" % agentName)
        self.log.info("Polling Time: %s" % pollingTime)
        self.log.info("Current Log File location: %s" % currentLogLocation)

        res = self.getLastAccessTime(currentLogLocation)
        if not res["OK"]:
            return res

        age = res["Value"]
        self.log.info("Current log file for %s is %d minutes old" % (agentName, (age.seconds / MINUTES)))

        maxLogAge = max(pollingTime + HOUR, 2 * HOUR)
        if age.seconds < maxLogAge:
            return S_OK()

        self.log.info("Current log file is too old for Agent %s" % agentName)
        self.accounting[agentName]["LogAge"] = age.seconds / MINUTES

        res = self.restartInstance(int(pid), agentName, self.restartAgents)
        if not res["OK"]:
            return res
        if res["Value"] != NO_RESTART:
            self.accounting[agentName]["Treatment"] = "Successfully Restarted"
            self.log.info("Agent %s has been successfully restarted" % agentName)

        return S_OK()

    def checkExecutor(self, executor, options):
        """Check the age of executor log file, if too old check for jobs in checking status, then restart the executors."""
        currentLogLocation = options["LogFileLocation"]
        pid = options["PID"]
        self.log.info("Checking executor: %s" % executor)
        self.log.info("Current Log File location: %s" % currentLogLocation)

        res = self.getLastAccessTime(currentLogLocation)
        if not res["OK"]:
            return res

        age = res["Value"]
        self.log.info("Current log file for %s is %d minutes old" % (executor, (age.seconds / MINUTES)))

        if age.seconds < 2 * HOUR:
            return S_OK()

        self.log.info("Current log file is too old for Executor %s" % executor)
        self.accounting[executor]["LogAge"] = age.seconds / MINUTES

        res = self.checkForCheckingJobs(executor)
        if not res["OK"]:
            return res
        if res["OK"] and res["Value"] == NO_CHECKING_JOBS:
            self.accounting.pop(executor, None)
            return S_OK(NO_RESTART)

        res = self.restartInstance(int(pid), executor, self.restartExecutors)
        if not res["OK"]:
            return res
        elif res["OK"] and res["Value"] != NO_RESTART:
            self.accounting[executor]["Treatment"] = "Successfully Restarted"
            self.log.info("Executor %s has been successfully restarted" % executor)

        return S_OK()

    def checkForCheckingJobs(self, executorName):
        """Check if there are checking jobs with the **executorName** as current MinorStatus."""
        attrDict = {"Status": "Checking", "MinorStatus": executorName}

        # returns list of jobs IDs
        resJobs = self.jobMonClient.getJobs(attrDict)
        if not resJobs["OK"]:
            self.logError("Could not get jobs for this executor", "%s: %s" % (executorName, resJobs["Message"]))
            return resJobs
        if resJobs["Value"]:
            self.log.info('Found %d jobs in "Checking" status for %s' % (len(resJobs["Value"]), executorName))
            return S_OK(CHECKING_JOBS)
        self.log.info('Found no jobs in "Checking" status for %s' % executorName)
        return S_OK(NO_CHECKING_JOBS)

    def componentControl(self):
        """Monitor and control component status as defined in the CS.

        Check for running and stopped components and ensure they have the proper status as defined in the CS
        Registry/Hosts/_HOST_/[Running|Stopped] sections

        :returns: :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_OK`,
           :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_ERROR`
        """
        # get the current status of the components

        resCurrent = self._getCurrentComponentStatus()
        if not resCurrent["OK"]:
            return resCurrent
        currentStatus = resCurrent["Value"]

        resDefault = self._getDefaultComponentStatus()
        if not resDefault["OK"]:
            return resDefault
        defaultStatus = resDefault["Value"]

        # ensure instances are in the right state
        shouldBe = {}
        shouldBe["Run"] = defaultStatus["Run"].intersection(currentStatus["Down"])
        shouldBe["Down"] = defaultStatus["Down"].intersection(currentStatus["Run"])
        shouldBe["Unknown"] = defaultStatus["All"].symmetric_difference(currentStatus["All"])

        self._ensureComponentRunning(shouldBe["Run"])
        self._ensureComponentDown(shouldBe["Down"])

        for instance in shouldBe["Unknown"]:
            self.logError("Unknown instance", "%r, either uninstall or add to config" % instance)

        return S_OK()

    def _getCurrentComponentStatus(self):
        """Get current status for components."""
        resOverall = self.sysAdminClient.getOverallStatus()
        if not resOverall["OK"]:
            return resOverall
        currentStatus = {"Down": set(), "Run": set(), "All": set()}
        informationDict = resOverall["Value"]
        for systemsDict in informationDict.values():
            for system, instancesDict in systemsDict.items():
                for instanceName, instanceInfoDict in instancesDict.items():
                    identifier = "%s__%s" % (system, instanceName)
                    runitStatus = instanceInfoDict.get("RunitStatus")
                    if runitStatus in ("Run", "Down"):
                        currentStatus[runitStatus].add(identifier)

        currentStatus["All"] = currentStatus["Run"] | currentStatus["Down"]
        return S_OK(currentStatus)

    def _getDefaultComponentStatus(self):
        """Get the configured status of the components."""
        host = socket.getfqdn()
        defaultStatus = {"Down": set(), "Run": set(), "All": set()}
        resRunning = gConfig.getOptionsDict(Path.cfgPath("/Registry/Hosts/", host, "Running"))
        resStopped = gConfig.getOptionsDict(Path.cfgPath("/Registry/Hosts/", host, "Stopped"))
        if not resRunning["OK"]:
            return resRunning
        if not resStopped["OK"]:
            return resStopped
        defaultStatus["Run"] = set(resRunning["Value"])
        defaultStatus["Down"] = set(resStopped["Value"])
        defaultStatus["All"] = defaultStatus["Run"] | defaultStatus["Down"]

        if defaultStatus["Run"].intersection(defaultStatus["Down"]):
            self.logError("Overlap in configuration", str(defaultStatus["Run"].intersection(defaultStatus["Down"])))
            return S_ERROR("Bad host configuration")

        return S_OK(defaultStatus)

    def _ensureComponentRunning(self, shouldBeRunning):
        """Ensure the correct components are running."""
        for instance in shouldBeRunning:
            self.log.info("Starting instance %s" % instance)
            system, name = instance.split("__")
            if self.controlComponents:
                res = self.sysAdminClient.startComponent(system, name)
                if not res["OK"]:
                    self.logError("Failed to start component:", "%s: %s" % (instance, res["Message"]))
                else:
                    self.accounting[instance]["Treatment"] = "Instance was down, started instance"
            else:
                self.accounting[instance]["Treatment"] = "Instance is down, should be started"

    def _ensureComponentDown(self, shouldBeDown):
        """Ensure the correct components are not running."""
        for instance in shouldBeDown:
            self.log.info("Stopping instance %s" % instance)
            system, name = instance.split("__")
            if self.controlComponents:
                res = self.sysAdminClient.stopComponent(system, name)
                if not res["OK"]:
                    self.logError("Failed to stop component:", "%s: %s" % (instance, res["Message"]))
                else:
                    self.accounting[instance]["Treatment"] = "Instance was running, stopped instance"
            else:
                self.accounting[instance]["Treatment"] = "Instance is running, should be stopped"

    def checkURLs(self):
        """Ensure that the running services have their URL in the Config."""
        self.log.info("Checking URLs")
        # get services again, in case they were started/stop in controlComponents
        gConfig.forceRefresh(fromMaster=True)

        # get port used for https based services
        try:
            tornadoSystemInstance = PathFinder.getSystemInstance(
                system="Tornado",
                setup=self.setup,
            )
            self._tornadoPort = gConfig.getValue(
                Path.cfgPath("/System/Tornado/", tornadoSystemInstance, "Port"),
                self._tornadoPort,
            )
        except RuntimeError:
            pass

        self.log.debug("Using Tornado Port:", self._tornadoPort)

        res = self.getRunningInstances(instanceType="Services", runitStatus="All")
        if not res["OK"]:
            return S_ERROR("Failure to get running services")
        self.services = res["Value"]
        for service, options in sorted(self.services.items()):
            self.log.debug("Checking URL for %s with options %s" % (service, options))
            # ignore SystemAdministrator, does not have URLs
            if "SystemAdministrator" in service:
                continue
            self._checkServiceURL(service, options)

        if self.csAPI.csModified and self.commitURLs:
            self.log.info("Commiting changes to the CS")
            result = self.csAPI.commit()
            if not result["OK"]:
                self.logError("Commit to CS failed", result["Message"])
                return S_ERROR("Failed to commit to CS")
        return S_OK()

    def _checkServiceURL(self, serviceName, options):
        """Ensure service URL is properly configured in the CS."""
        url = self._getURL(serviceName, options)
        system = options["System"]
        module = options["Module"]
        self.log.info("Checking URLs for %s/%s" % (system, module))
        urlsConfigPath = Path.cfgPath(PathFinder.getSystemURLSection(system=system, setup=self.setup), module)
        urls = gConfig.getValue(urlsConfigPath, [])
        self.log.debug("Found configured URLs for %s: %s" % (module, urls))
        self.log.debug("This URL is %s" % url)
        runitStatus = options["RunitStatus"]
        wouldHave = "Would have " if not self.commitURLs else ""
        if runitStatus == "Run" and url not in urls:
            urls.append(url)
            message = "%sAdded URL %s to URLs for %s/%s" % (wouldHave, url, system, module)
            self.log.info(message)
            self.accounting[serviceName + "/URL"]["Treatment"] = message
            self.csAPI.modifyValue(urlsConfigPath, ",".join(urls))
        if runitStatus == "Down" and url in urls:
            urls.remove(url)
            message = "%sRemoved URL %s from URLs for %s/%s" % (wouldHave, url, system, module)
            self.log.info(message)
            self.accounting[serviceName + "/URL"]["Treatment"] = message
            self.csAPI.modifyValue(urlsConfigPath, ",".join(urls))

    def _getURL(self, serviceName, options):
        """Return URL for the service."""
        system = options["System"]
        port = options.get("Port", self._tornadoPort)
        host = socket.getfqdn()
        protocol = options.get("Protocol", "dips")
        url = "%s://%s:%s/%s/%s" % (protocol, host, port, system, serviceName)
        return url



#! /usr/bin/env python
"""
Get Pilots Logging for specific Pilot UUID or Job ID.

WARNING: Only one option (either uuid or jobid) should be used.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

uuid = None
jobid = None


def setUUID(optVal):
    """
    Set UUID from arguments
    """
    global uuid
    uuid = optVal
    return S_OK()


def setJobID(optVal):
    """
    Set JobID from arguments
    """
    global jobid
    jobid = optVal
    return S_OK()


@Script()
def main():
    global uuid
    global jobid
    Script.registerSwitch("u:", "uuid=", "get PilotsLogging for given Pilot UUID", setUUID)
    Script.registerSwitch("j:", "jobid=", "get PilotsLogging for given Job ID", setJobID)
    Script.parseCommandLine()

    from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient

    if jobid:
        result = PilotManagerClient().getPilots(jobid)
        if not result["OK"]:
            gLogger.error(result["Message"])
            DIRAC.exit(1)
        gLogger.debug(result["Value"])
        uuid = list(result["Value"])[0]

    result = PilotManagerClient().getPilotLoggingInfo(uuid)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRAC.exit(1)
    gLogger.notice(result["Value"])

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" TSCatalogClient class represents the Transformation Manager service
    as a DIRAC Catalog service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase


class TSCatalogClient(FileCatalogClientBase):

    """Exposes the catalog functionality available in the DIRAC/TransformationHandler"""

    # List of common File Catalog methods implemented by this client
    WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + ["addFile", "removeFile", "setMetadata"]

    NO_LFN_METHODS = ["setMetadata"]

    def __init__(self, url=None, **kwargs):

        self.serverURL = "Transformation/TransformationManager" if not url else url
        super(TSCatalogClient, self).__init__(self.serverURL, **kwargs)

    @checkCatalogArguments
    def addFile(self, lfns, force=False):
        rpcClient = self._getRPC()
        return rpcClient.addFile(lfns, force)

    @checkCatalogArguments
    def removeFile(self, lfns):
        rpcClient = self._getRPC()
        successful = {}
        failed = {}
        listOfLists = breakListIntoChunks(lfns, 100)
        for fList in listOfLists:
            res = rpcClient.removeFile(fList)
            if not res["OK"]:
                return res
            successful.update(res["Value"]["Successful"])
            failed.update(res["Value"]["Failed"])
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def setMetadata(self, path, metadatadict):
        """Set metadata parameter for the given path

        :return: Successful/Failed dict.
        """
        rpcClient = self._getRPC()
        return rpcClient.setMetadata(path, metadatadict)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os

from DIRAC import gLogger
from DIRAC.Core.Security import Locations
from DIRAC.Core.Utilities.ReturnValues import S_ERROR, S_OK
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security.X509Certificate import X509Certificate  # pylint: disable=import-error


# Eventhough SSLTransport is not used in this file, it is imported in other module from there,
# so do not remove these imports !
from DIRAC.Core.DISET.private.Transports.M2SSLTransport import SSLTransport


def delegate(delegationRequest, kwargs):
    """
    Check delegate!
    """
    if kwargs.get("useCertificates"):
        chain = X509Chain()
        certTuple = Locations.getHostCertificateAndKeyLocation()
        chain.loadChainFromFile(certTuple[0])
        chain.loadKeyFromFile(certTuple[1])
    elif "proxyObject" in kwargs:
        chain = kwargs["proxyObject"]
    else:
        if "proxyLocation" in kwargs:
            procLoc = kwargs["proxyLocation"]
        else:
            procLoc = Locations.getProxyLocation()
        chain = X509Chain()
        chain.loadChainFromFile(procLoc)
        chain.loadKeyFromFile(procLoc)
    return chain.generateChainFromRequestString(delegationRequest)


def checkSanity(urlTuple, kwargs):
    """
    Check that all ssl environment is ok
    """
    useCerts = False
    certFile = ""
    if "useCertificates" in kwargs and kwargs["useCertificates"]:
        certTuple = Locations.getHostCertificateAndKeyLocation()
        if not certTuple:
            gLogger.error("No cert/key found! ")
            return S_ERROR("No cert/key found! ")
        certFile = certTuple[0]
        useCerts = True
    elif "proxyString" in kwargs:
        if not isinstance(kwargs["proxyString"], six.string_types if six.PY2 else bytes):
            gLogger.error("proxyString parameter is not a valid type", str(type(kwargs["proxyString"])))
            return S_ERROR("proxyString parameter is not a valid type")
    else:
        if "proxyLocation" in kwargs:
            certFile = kwargs["proxyLocation"]
        else:
            certFile = Locations.getProxyLocation()
        if not certFile:
            gLogger.error("No proxy found")
            return S_ERROR("No proxy found")
        elif not os.path.isfile(certFile):
            gLogger.error("Proxy file does not exist", certFile)
            return S_ERROR("%s proxy file does not exist" % certFile)

    # For certs always check CA's. For clients skipServerIdentityCheck
    if "skipCACheck" not in kwargs or not kwargs["skipCACheck"]:
        if not Locations.getCAsLocation():
            gLogger.error("No CAs found!")
            return S_ERROR("No CAs found!")

    if "proxyString" in kwargs:
        certObj = X509Chain()
        retVal = certObj.loadChainFromString(kwargs["proxyString"])
        if not retVal["OK"]:
            gLogger.error("Can't load proxy string")
            return S_ERROR("Can't load proxy string")
    else:
        if useCerts:
            certObj = X509Certificate()
            certObj.loadFromFile(certFile)
        else:
            certObj = X509Chain()
            certObj.loadChainFromFile(certFile)

    retVal = certObj.hasExpired()
    if not retVal["OK"]:
        gLogger.error("Can't verify proxy or certificate file", "%s:%s" % (certFile, retVal["Message"]))
        return S_ERROR("Can't verify file %s:%s" % (certFile, retVal["Message"]))
    else:
        if retVal["Value"]:
            notAfter = certObj.getNotAfterDate()
            if notAfter["OK"]:
                notAfter = notAfter["Value"]
            else:
                notAfter = "unknown"
            gLogger.error("PEM file has expired", "%s is not valid after %s" % (certFile, notAfter))
            return S_ERROR("PEM file %s has expired, not valid after %s" % (certFile, notAfter))

    idDict = {}
    retVal = certObj.getDIRACGroup(ignoreDefault=True)
    if retVal["OK"] and retVal["Value"] is not False:
        idDict["group"] = retVal["Value"]
    if useCerts:
        idDict["DN"] = certObj.getSubjectDN()["Value"]
    else:
        idDict["DN"] = certObj.getIssuerCert()["Value"].getSubjectDN()["Value"]

    return S_OK(idDict)



""" DIRAC Graphs package provides tools for creation of various plots to provide
    graphical representation of the DIRAC Monitoring and Accounting data

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from os.path import dirname, join

# Make sure the the Agg backend is used despite arbitrary configuration
import matplotlib
import six

matplotlib.use("agg")

from matplotlib.pyplot import hist

import DIRAC

from DIRAC.Core.Utilities.Graphs.Graph import Graph
from DIRAC.Core.Utilities.Graphs.GraphUtilities import evalPrefs

common_prefs = {
    "background_color": "white",
    "figure_padding": 12,
    "plot_grid": "1:1",
    "plot_padding": 0,
    "frame": "On",
    "font": "DejaVu Sans",
    "font_family": "sans-serif",
    "dpi": 100,
    "legend": True,
    "legend_position": "bottom",
    "legend_max_rows": 99,
    "legend_max_columns": 4,
    "square_axis": False,
    "scale_data": None,
    "scale_ticks": None,
}

graph_large_prefs = {
    "width": 1000,
    "height": 700,
    "text_size": 8,
    "subtitle_size": 10,
    "subtitle_padding": 5,
    "title_size": 15,
    "title_padding": 5,
    "text_padding": 5,
    "figure_padding": 15,
    "plot_title_size": 12,
    "legend_width": 980,
    "legend_height": 150,
    "legend_padding": 20,
    "limit_labels": 15,
    "graph_time_stamp": True,
}

graph_normal_prefs = {
    "width": 800,
    "height": 600,
    "text_size": 8,
    "subtitle_size": 10,
    "subtitle_padding": 5,
    "title_size": 15,
    "title_padding": 10,
    "text_padding": 5,
    "figure_padding": 12,
    "plot_title_size": 12,
    "legend_width": 780,
    "legend_height": 120,
    "legend_padding": 20,
    "limit_labels": 15,
    "graph_time_stamp": True,
    "label_text_size": 14,
}

graph_small_prefs = {
    "width": 450,
    "height": 330,
    "text_size": 10,
    "subtitle_size": 5,
    "subtitle_padding": 4,
    "title_size": 10,
    "title_padding": 6,
    "text_padding": 3,
    "figure_padding": 10,
    "plot_title_size": 8,
    "legend_width": 430,
    "legend_height": 50,
    "legend_padding": 10,
    "limit_labels": 15,
    "graph_time_stamp": True,
}

graph_thumbnail_prefs = {
    "width": 100,
    "height": 80,
    "text_size": 6,
    "subtitle_size": 0,
    "subtitle_padding": 0,
    "title_size": 8,
    "title_padding": 2,
    "text_padding": 1,
    "figure_padding": 2,
    "plot_title": "NoTitle",
    "legend": False,
    "plot_axis_grid": False,
    "plot_axis": False,
    "plot_axis_labels": False,
    "graph_time_stamp": False,
    "tight_bars": True,
}


def graph(data, fileName, *args, **kw):

    prefs = evalPrefs(*args, **kw)
    graph_size = prefs.get("graph_size", "normal")

    if graph_size == "normal":
        defaults = graph_normal_prefs
    elif graph_size == "small":
        defaults = graph_small_prefs
    elif graph_size == "thumbnail":
        defaults = graph_thumbnail_prefs
    elif graph_size == "large":
        defaults = graph_large_prefs

    graph = Graph()
    graph.makeGraph(data, common_prefs, defaults, prefs)
    graph.writeGraph(fileName, "PNG")
    return DIRAC.S_OK({"plot": fileName})


def __checkKW(kw):
    if "watermark" not in kw:
        if six.PY2:
            kw["watermark"] = join(dirname(DIRAC.__file__), "Core/Utilities/Graphs/Dwatermark.png")
        else:
            from importlib.resources import files  # pylint: disable=no-name-in-module,import-error

            kw["watermark"] = str(files("DIRAC.Core.Utilities.Graphs") / "Dwatermark.png")
    return kw


def barGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    graph(data, fileName, plot_type="BarGraph", statistics_line=True, *args, **kw)


def lineGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    graph(data, fileName, plot_type="LineGraph", statistics_line=True, *args, **kw)


def curveGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    graph(data, fileName, plot_type="CurveGraph", statistics_line=False, *args, **kw)


def cumulativeGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    graph(data, fileName, plot_type="LineGraph", cumulate_data=True, *args, **kw)


def pieGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    prefs = {"xticks": False, "yticks": False, "legend_position": "right"}
    graph(data, fileName, prefs, plot_type="PieGraph", *args, **kw)


def qualityGraph(data, fileName, *args, **kw):
    kw = __checkKW(kw)
    prefs = {"plot_axis_grid": False}
    graph(data, fileName, prefs, plot_type="QualityMapGraph", *args, **kw)


def textGraph(text, fileName, *args, **kw):
    kw = __checkKW(kw)
    prefs = {"text_image": text}
    graph({}, fileName, prefs, *args, **kw)


def histogram(data, fileName, bins=None, *args, **kw):
    kw = __checkKW(kw)
    if bins is None:
        bins = "auto"
    histograms = None
    spans = []
    if isinstance(data, list):
        if isinstance(data[0], dict):
            for plots in data:
                for plot in plots:
                    values, vbins, _patches = hist(plots[plot], bins)
                    # we can have multiple plots in a figure, although we
                    # are not using it, but still it is good to have in histograms as well.
                    if "plot_grid" in kw and kw["plot_grid"] not in "1:1":
                        if not histograms:
                            histograms = []
                        histograms.append({plot: dict(zip(vbins, values))})
                    else:
                        if not histograms:
                            histograms = {}
                        histograms[plot] = dict(zip(vbins, values))
                    if vbins.size:
                        bins = len(vbins)
                        spans.append((max(plots[plot]) - min(plots[plot]) + 0.1) / float(bins) * 0.95)
        else:
            values, vbins, _patches = hist(data, bins)
            histograms = dict(zip(vbins, values))
            if vbins.size:
                bins = len(vbins)
            spans = [(max(data) - min(data)) / float(bins) * 0.95]
    kw = __checkKW(kw)
    span = max(spans)
    graph(histograms, fileName, plot_type="BarGraph", span=span, statistics_line=True, *args, **kw)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC.Core.Workflow.Step import *
from DIRAC.Core.Workflow.Workflow import *

bodyTestApp = """class TestAppModule:

    def __init__(self):
        pass

    def initialize(self,name,version,parameters):
        print( 'Test application initialization %s %s' % ( name, str( version ) ) )
        return 1

    def execute(self):
        if not self.enable:
            print 'Module ',str(type(self)),'is disabled'
            return 1
        print( 'Test application execution' )
        print 'Module MODULE.getDescrShort() =', self.MODULE.getDescrShort()
        print 'Module MODULE_DEFINITION_NAME =', self.MODULE_DEFINITION_NAME
        print 'Module MODULE_INSTANCE_NAME =', self.MODULE_INSTANCE_NAME
        print 'Module STEP.getDescrShort() =', self.STEP.getDescrShort()
        print 'Module STEP_DEFINITION_NAME =', self.STEP_DEFINITION_NAME
        print 'Module STEP_INSTANCE_NAME =', self.STEP_INSTANCE_NAME

        return 1"""

md1 = ModuleDefinition("TestAppModule")
md1.addParameter(Parameter("enable", "True", "bool", "", "", True, False, "If False execution is disabled"))
md1.setBody(bodyTestApp)
md1.setDescrShort("Empty Module")
md1.setDescription("Empty Module to do some testing")
md1.setRequired("")
md1.setOrigin("")
md1.setVersion(0.1)

sd1 = StepDefinition("TestAppStep")
sd1.addModule(md1)
mi1 = sd1.createModuleInstance("TestAppModule", "testappmod1")
wf1 = Workflow("TestAppWF")
wf1.addStep(sd1)
si1 = wf1.createStepInstance("TestAppStep", "testappstep1")
print(wf1.createCode())
# eval(compile(wf1.createCode(),'<string>','exec'))
wf1.execute()



"""
This class is used to define the plot using the plot attributes.
"""

from DIRAC import S_OK, gLogger

from DIRAC.MonitoringSystem.Client.Types.RMSMonitoring import RMSMonitoring
from DIRAC.MonitoringSystem.private.Plotters.BasePlotter import BasePlotter

__RCSID__ = "$Id$"


class RMSMonitoringPlotter(BasePlotter):

    """
    .. class:: RMSMonitoringPlotter

    It is used to crate the plots.

    param: str _typeName monitoring type
    param: list _typeKeyFields list of keys what we monitor (list of attributes)
    """

    _typeName = "RMSMonitoring"
    _typeKeyFields = RMSMonitoring().keyFields

    def __reportAllResources(self, reportRequest, metric, unit):

        retVal = self._getTimedData(
            startTime=reportRequest["startTime"],
            endTime=reportRequest["endTime"],
            selectField=metric,
            preCondDict=reportRequest["condDict"],
            metadataDict={"DynamicBucketing": False, "metric": "sum"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        try:
            _, _, _, unitName = self._findSuitableUnit(dataDict, self._getAccumulationMaxValue(dataDict), unit)
        except AttributeError as e:
            gLogger.warn(e)
            unitName = unit

        return S_OK({"data": dataDict, "granularity": granularity, "unit": unitName})

    def __plotAllResources(self, reportRequest, plotInfo, filename, title):

        metadata = {
            "title": "%s by %s" % (title, reportRequest["grouping"]),
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "skipEdgeColor": True,
            "ylabel": plotInfo["unit"],
        }

        plotInfo["data"] = self._fillWithZero(
            granularity=plotInfo["granularity"],
            startEpoch=reportRequest["startTime"],
            endEpoch=reportRequest["endTime"],
            dataDict=plotInfo["data"],
        )

        return self._generateStackedLinePlot(filename=filename, dataDict=plotInfo["data"], metadata=metadata)

    _reportnbObjectName = "Number of objects(Request, Operation, or File)"

    def _reportnbObject(self, reportRequest):
        """It is used to retrieve the data from the database.

        :param dict reportRequest: contains attributes used to create the plot.
        :return S_OK or S_ERROR {'data':value1, 'granularity':value2} value1 is a dictionary, value2 is the bucket length
        """
        return self.__reportAllResources(reportRequest, "nbObject", "objects")

    def _plotnbObject(self, reportRequest, plotInfo, filename):
        """It creates the plot.

        :param dict reportRequest: plot attributes
        :param dict plotInfo: contains all the data which are used to create the plot
        :param str filename:
        :return: S_OK or S_ERROR { 'plot' : value1, 'thumbnail' : value2 } value1 and value2 are TRUE/FALSE
        """
        return self.__plotAllResources(
            reportRequest, plotInfo, filename, "Number of objects(Request, Operation, or File)"
        )



""" The WorkflowTaskAgent takes workflow tasks created in the TransformationDB and submits them to the WMS.

The following options can be set for the WorkflowTaskAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN WorkflowTaskAgent
  :end-before: ##END
  :dedent: 2
  :caption: WorkflowTaskAgent options

The options *SubmitTasks*, *MonitorTasks*, *MonitorFiles*, and *CheckReserved*
need to be assigned any non-empty value to be activated

* .. versionadded:: v6r20p5

   It is possible to run the RequestTaskAgent without a *shifterProxy* or
   *ShifterCredentials*, in this case the credentials of the authors of the
   transformations are used to submit the jobs to the RMS. This enables the use of
   a single RequestTaskAgent for multiple VOs. See also the section about the
   :ref:`trans-multi-vo`.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Agent.TaskManagerAgentBase import TaskManagerAgentBase

AGENT_NAME = "Transformation/WorkflowTaskAgent"


class WorkflowTaskAgent(TaskManagerAgentBase):
    """An AgentModule class to submit workflow tasks"""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        TaskManagerAgentBase.__init__(self, *args, **kwargs)

        self.transType = []

    def initialize(self):
        """Standard initialize method"""
        res = TaskManagerAgentBase.initialize(self)
        if not res["OK"]:
            return res

        agentTSTypes = self.am_getOption("TransType", [])
        if agentTSTypes:
            self.transType = agentTSTypes
        else:
            self.transType = Operations().getValue("Transformations/DataProcessing", ["MCSimulation", "Merge"])

        return S_OK()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import datetime
import errno
import json
from sqlalchemy import orm

from DIRAC.DataManagementSystem.Client.FTS3Job import FTS3Job
from DIRAC.DataManagementSystem.private import FTS3Utilities
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.FrameworkSystem.Client.Logger import gLogger

from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.DErrno import cmpError


from DIRAC import S_OK, S_ERROR

from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
from DIRAC.DataManagementSystem.Client.FTS3File import FTS3File
from DIRAC.Core.Utilities.JEncode import JSerializable

from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Operation import Operation as rmsOperation
from DIRAC.RequestManagementSystem.Client.File import File as rmsFile
from DIRAC.RequestManagementSystem.Client.Request import Request as rmsRequest

from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup


class FTS3Operation(JSerializable):
    """Abstract class to represent an operation to be executed by FTS. It is a
    container for FTSFiles, as well as for FTSJobs.

    There can be a mapping between one FTS3Operation and one RMS Operation.

    The FTS3Operation takes care of generating the appropriate FTSJobs,
    and to perform a callback when the work with FTS is over. The actual
    generation and callback depends on the subclass.

    This class should not be instantiated directly, but rather one of its
    subclass
    """

    # START states
    ALL_STATES = [
        "Active",  # Default state until FTS has done everything
        "Processed",  # Interactions with FTS done, but callback not done
        "Finished",  # Everything was done
        "Canceled",  # Canceled by the user
        "Failed",  # I don't know yet
    ]
    FINAL_STATES = ["Finished", "Canceled", "Failed"]
    INIT_STATE = "Active"
    # END states

    _attrToSerialize = [
        "operationID",
        "username",
        "userGroup",
        "rmsReqID",
        "rmsOpID",
        "sourceSEs",
        "ftsFiles",
        "activity",
        "priority",
        "ftsJobs",
        "creationTime",
        "lastUpdate",
        "error",
        "status",
    ]

    def __init__(
        self,
        ftsFiles=None,
        username=None,
        userGroup=None,
        rmsReqID=-1,
        rmsOpID=0,
        sourceSEs=None,
        activity=None,
        priority=None,
    ):
        """
        :param ftsFiles: list of FTS3Files object that belongs to the operation
        :param username: username whose proxy should be used
        :param userGroup: group that should be used with username
        :param rmsReqID: ID of the Request in the RMS system
        :param rmsOpID: ID of the Operation in the RMS system
        :param sourceSEs: list of SE to be used as source (if applicable)
        :param activity: FTS activity to use
        :param priority: FTS priority to use

        """
        ############################
        # persistent attributes

        self.username = username
        self.userGroup = userGroup

        self.rmsReqID = rmsReqID
        self.rmsOpID = rmsOpID

        if isinstance(sourceSEs, list):
            sourceSEs = ",".join(sourceSEs)

        self.sourceSEs = sourceSEs

        self.ftsFiles = ftsFiles if ftsFiles else []

        self.activity = activity
        self.priority = priority

        self.ftsJobs = []

        now = datetime.datetime.utcnow().replace(microsecond=0)

        self.creationTime = now
        self.lastUpdate = now
        self.error = None
        self.status = FTS3Operation.INIT_STATE

        ########################

        self.reqClient = None
        self.dManager = None
        self._log = None
        self.fts3Plugin = None
        self.init_on_load()

    @orm.reconstructor
    def init_on_load(self):
        """This method initializes some attributes.
        It is called by sqlalchemy (which does not call __init__)

        """
        self._vo = None

        # Note that in the case of an FTS3Operation created from an RMS
        # object, the members here will probably be "wrong" in the sense
        # that the VO will not be known by then.
        # It does not really matter however, since we do not perform anything
        # on an operation created this way, it's just to be then serialized
        # in the DB.
        self.dManager = DataManager()
        self.rssClient = ResourceStatus()
        self.fts3Plugin = FTS3Utilities.getFTS3Plugin(vo=self.vo)

        opID = getattr(self, "operationID", None)
        loggerName = "%s/" % opID if opID else ""
        loggerName += "req_%s/op_%s" % (self.rmsReqID, self.rmsOpID)

        self._log = gLogger.getSubLogger(loggerName)

    @property
    def vo(self):
        """:returns: return vo of the usergroup"""
        if self._vo:
            return self._vo

        if self.userGroup:
            self._vo = getVOForGroup(self.userGroup)

        return self._vo

    def isTotallyProcessed(self):
        """Returns True if and only if there is nothing
        else to be done by FTS for this operation.
        All files are successful or definitely failed
        """

        if self.status == "Processed":
            return True

        fileStatuses = set([f.status for f in self.ftsFiles])

        # If all the files are in a final state
        if fileStatuses <= set(FTS3File.FINAL_STATES):
            self.status = "Processed"
            return True

        return False

    def _getFilesToSubmit(self, maxAttemptsPerFile=10):
        """Return the list of FTS3files that can be submitted
        Either because they never were submitted, or because
        we can make more attempts

        :param maxAttemptsPerFile: the maximum number of attempts to be tried for a file

        :return: List of FTS3File to submit
        """

        toSubmit = []

        for ftsFile in self.ftsFiles:
            if ftsFile.attempt >= maxAttemptsPerFile:
                ftsFile.status = "Defunct"
            # The file was never submitted or
            # The file failed from the point of view of FTS
            # but no more than the maxAttemptsPerFile
            elif ftsFile.status in [FTS3File.INIT_STATE] + FTS3File.FTS_FAILED_STATES:
                toSubmit.append(ftsFile)

        return toSubmit

    @staticmethod
    def _checkSEAccess(seName, accessType, vo=None):
        """Check the Status of a storage element

        :param seName: name of the StorageElement
        :param accessType ReadAccess, WriteAccess,CheckAccess,RemoveAccess

        :return: S_ERROR if not allowed or error, S_OK() otherwise
        """
        # Check that the target is writable
        # access = self.rssClient.getStorageElementStatus( seName, accessType )
        # if not access["OK"]:
        #   return access
        # if access["Value"][seName][accessType] not in ( "Active", "Degraded" ):
        #   return S_ERROR( "%s does not have %s in Active or Degraded" % ( seName, accessType ) )

        status = StorageElement(seName, vo=vo).getStatus()
        if not status["OK"]:
            return status

        status = status["Value"]

        accessType = accessType.replace("Access", "")
        if not status[accessType]:
            return S_ERROR(errno.EACCES, "%s does not have %s in Active or Degraded" % (seName, accessType))

        return S_OK()

    def _createNewJob(self, jobType, ftsFiles, targetSE, sourceSE=None):
        """Create a new FTS3Job object

        :param jobType: type of job to create (Transfer, Staging, Removal)
        :param ftsFiles: list of FTS3File objects the job has to work on
        :param targetSE: SE on which to operate
        :param sourceSE: source SE, only useful for Transfer jobs

        :return: FTS3Job object
        """

        newJob = FTS3Job()
        newJob.type = jobType
        newJob.sourceSE = sourceSE
        newJob.targetSE = targetSE
        newJob.activity = self.activity
        newJob.priority = self.priority
        newJob.username = self.username
        newJob.userGroup = self.userGroup
        newJob.vo = self.vo
        newJob.filesToSubmit = ftsFiles
        newJob.operationID = getattr(self, "operationID")
        newJob.rmsReqID = self.rmsReqID

        return newJob

    def _callback(self):
        """Actually performs the callback"""
        raise NotImplementedError("You should not be using the base class")

    def callback(self):
        """Trigger the callback once all the FTS interactions are done
        and update the status of the Operation to 'Finished' if successful
        """
        self.reqClient = ReqClient()

        res = self._callback()

        if res["OK"]:
            self.status = "Finished"

        return res

    def prepareNewJobs(self, maxFilesPerJob=100, maxAttemptsPerFile=10):
        """Prepare the new jobs that have to be submitted

        :param maxFilesPerJob: maximum number of files assigned to a job
        :param maxAttemptsPerFile: maximum number of retry after an fts failure

        :return: list of jobs
        """
        raise NotImplementedError("You should not be using the base class")

    def _updateRmsOperationStatus(self):
        """Update the status of the Files in the rms operation
        :return: S_OK with a dict:
                      * request: rms Request object
                      * operation: rms Operation object
                      * ftsFilesByTarget: dict {SE: [ftsFiles that were successful]}
        """

        log = self._log.getLocalSubLogger(
            "_updateRmsOperationStatus/%s/%s" % (getattr(self, "operationID"), self.rmsReqID)
        )

        res = self.reqClient.getRequest(self.rmsReqID)
        if not res["OK"]:
            return res

        request = res["Value"]

        res = request.getWaiting()

        if not res["OK"]:
            log.error("Unable to find 'Scheduled' operation in request")
            res = self.reqClient.putRequest(request, useFailoverProxy=False, retryMainService=3)
            if not res["OK"]:
                log.error("Could not put back the request !", res["Message"])
            return S_ERROR("Could not find scheduled operation")

        operation = res["Value"]

        # We index the files of the operation by their IDs
        rmsFileIDs = {}

        for opFile in operation:
            rmsFileIDs[opFile.FileID] = opFile

        # Files that failed to transfer
        defunctRmsFileIDs = set()

        # { SE : [FTS3Files] }
        ftsFilesByTarget = {}
        for ftsFile in self.ftsFiles:

            if ftsFile.status == "Defunct":
                log.info(
                    "File failed to transfer, setting it to failed in RMS", "%s %s" % (ftsFile.lfn, ftsFile.targetSE)
                )
                defunctRmsFileIDs.add(ftsFile.rmsFileID)
                continue

            if ftsFile.status == "Canceled":
                log.info("File canceled, setting it Failed in RMS", "%s %s" % (ftsFile.lfn, ftsFile.targetSE))
                defunctRmsFileIDs.add(ftsFile.rmsFileID)
                continue

            # SHOULD NEVER HAPPEN !
            if ftsFile.status != "Finished":
                log.error("Callback called with file in non terminal state", "%s %s" % (ftsFile.lfn, ftsFile.targetSE))
                res = self.reqClient.putRequest(request, useFailoverProxy=False, retryMainService=3)
                if not res["OK"]:
                    log.error("Could not put back the request !", res["Message"])
                return S_ERROR("Callback called with file in non terminal state")

            ftsFilesByTarget.setdefault(ftsFile.targetSE, []).append(ftsFile)

        # Now, we set the rmsFile as done in the operation, providing
        # that they are not in the defunctFiles.
        # We cannot do this in the previous list because in the FTS system,
        # each destination is a separate line in the DB but not in the RMS

        for ftsFile in self.ftsFiles:
            opFile = rmsFileIDs[ftsFile.rmsFileID]

            opFile.Status = "Failed" if ftsFile.rmsFileID in defunctRmsFileIDs else "Done"

        return S_OK({"request": request, "operation": operation, "ftsFilesByTarget": ftsFilesByTarget})

    @classmethod
    def fromRMSObjects(cls, rmsReq, rmsOp, username):
        """Construct an FTS3Operation object from the RMS Request and Operation corresponding.
        The attributes taken are the OwnerGroup, Request and Operation IDS, sourceSE,
        and activity and priority if they are defined in the Argument field of the operation

        :param rmsReq: RMS Request object
        :param rmsOp: RMS Operation object
        :param username: username to which associate the FTS3Operation (normally comes from the Req OwnerDN)

        :returns: FTS3Operation object
        """

        ftsOp = cls()
        ftsOp.username = username
        ftsOp.userGroup = rmsReq.OwnerGroup

        ftsOp.rmsReqID = rmsReq.RequestID
        ftsOp.rmsOpID = rmsOp.OperationID

        ftsOp.sourceSEs = rmsOp.SourceSE

        try:
            argumentDic = json.loads(rmsOp.Arguments)

            ftsOp.activity = argumentDic["activity"]
            ftsOp.priority = argumentDic["priority"]
        except Exception:
            pass

        return ftsOp


class FTS3TransferOperation(FTS3Operation):
    """Class to be used for a Replication operation"""

    def prepareNewJobs(self, maxFilesPerJob=100, maxAttemptsPerFile=10):

        log = self._log.getSubLogger("_prepareNewJobs")

        filesToSubmit = self._getFilesToSubmit(maxAttemptsPerFile=maxAttemptsPerFile)
        log.debug("%s ftsFiles to submit" % len(filesToSubmit))

        newJobs = []

        # {targetSE : [FTS3Files] }
        res = FTS3Utilities.groupFilesByTarget(filesToSubmit)
        if not res["OK"]:
            return res
        filesGroupedByTarget = res["Value"]

        for targetSE, ftsFiles in filesGroupedByTarget.items():

            res = self._checkSEAccess(targetSE, "WriteAccess", vo=self.vo)

            if not res["OK"]:
                # If the SE is currently banned, we just skip it
                if cmpError(res, errno.EACCES):
                    log.info("Write access currently not permitted to %s, skipping." % targetSE)
                else:
                    log.error(res)
                    for ftsFile in ftsFiles:
                        ftsFile.attempt += 1
                continue

            sourceSEs = self.sourceSEs.split(",") if self.sourceSEs is not None else []
            # { sourceSE : [FTSFiles] }
            res = FTS3Utilities.selectUniqueSource(ftsFiles, self.fts3Plugin, allowedSources=sourceSEs)

            if not res["OK"]:
                return res

            uniqueTransfersBySource, failedFiles = res["Value"]

            # Treat the errors of the failed files
            for ftsFile, errMsg in failedFiles.items():
                log.error("Error when selecting random sources", "%s, %s" % (ftsFile.lfn, errMsg))
                # If the error is that the file does not exist in the catalog
                # fail it !
                if cmpError(errMsg, errno.ENOENT):
                    log.error("The file does not exist, setting it Defunct", "%s" % ftsFile.lfn)
                    ftsFile.status = "Defunct"

            # We don't need to check the source, since it is already filtered by the DataManager
            for sourceSE, ftsFiles in uniqueTransfersBySource.items():

                if self.__needsMultiHopStaging(sourceSE, targetSE):
                    log.verbose("Needs multihop staging, max files per job is 1")
                    maxFilesPerJob = 1

                for ftsFilesChunk in breakListIntoChunks(ftsFiles, maxFilesPerJob):

                    newJob = self._createNewJob("Transfer", ftsFilesChunk, targetSE, sourceSE=sourceSE)

                    newJobs.append(newJob)

        return S_OK(newJobs)

    def __needsMultiHopStaging(self, sourceSEName, destSEName):
        """Checks whether transfers between the two SE given as parameters
        need a multi hop transfer to stage with a different protocol
        than the transfer one.

        :param str sourceSEName: source storage element name
        :param str destSEName: destination storage element name

        :returns: boolean
        """
        srcSE = StorageElement(sourceSEName, vo=self.vo)
        dstSE = StorageElement(destSEName, vo=self.vo)
        srcIsTape = srcSE.getStatus()["Value"].get("TapeSE", True)

        if not srcIsTape:
            return False

        # To know if we will need a multihop staging transfer,
        # we check whether we can generate transfer URLs
        # for a fake LFN, and see if the protocol we get
        # is compatible with staging
        tpcProtocols = self.fts3Plugin.selectTPCProtocols(sourceSEName=sourceSEName, destSEName=destSEName)

        res = dstSE.generateTransferURLsBetweenSEs("/%s/fakeLFN" % self.vo, srcSE, protocols=tpcProtocols)

        # There is an error, but let's ignore it,
        # it will be dealt with in the FTS3Job logic
        if not res["OK"]:
            return False

        srcProto, _destProto = res["Value"]["Protocols"]
        if srcProto not in srcSE.localStageProtocolList:
            return True

        return False

    def _callback(self):
        """ " After a Transfer operation, we have to update the matching Request in the
        RMS, and add the registration operation just before the ReplicateAndRegister one

        NOTE: we don't use ReqProxy when putting the request back to avoid operational hell
        """

        log = self._log.getSubLogger("callback")

        # In case there is no Request associated to the Transfer
        # we do not do the callback. Not really advised, but there is a feature
        # request to use the FTS3 system without RMS
        if self.rmsReqID == -1:
            return S_OK()

        # Now we check the status of the Request.
        # in principle, it should be scheduled
        res = self.reqClient.getRequestStatus(self.rmsReqID)
        if not res["OK"]:
            log.error("Could not get request status", res)
            return res
        status = res["Value"]

        # If it is not scheduled, something went wrong
        # and we will not modify it
        if status != "Scheduled":
            # If the Request is in a final state, just leave it,
            # and we consider our job done.
            # (typically happens when the callback had already been done but not persisted to the FTS3DB)
            if status in rmsRequest.FINAL_STATES:
                log.warn(
                    "Request with id %s is not Scheduled (%s), but okay it is in a Final State"
                    % (self.rmsReqID, status)
                )
                return S_OK()
            # If the Request is not in a final state, then something really wrong is going on,
            # and we do not do anything, keep ourselves pending
            else:
                return S_ERROR("Request with id %s is not Scheduled:%s" % (self.rmsReqID, status))

        res = self._updateRmsOperationStatus()

        if not res["OK"]:
            return res

        ftsFilesByTarget = res["Value"]["ftsFilesByTarget"]
        request = res["Value"]["request"]
        operation = res["Value"]["operation"]

        registrationProtocols = DMSHelpers(vo=self.vo).getRegistrationProtocols()

        log.info("will create %s 'RegisterReplica' operations" % len(ftsFilesByTarget))

        for target, ftsFileList in ftsFilesByTarget.items():
            log.info(
                "creating 'RegisterReplica' operation for targetSE %s with %s files..." % (target, len(ftsFileList))
            )
            registerOperation = rmsOperation()
            registerOperation.Type = "RegisterReplica"
            registerOperation.Status = "Waiting"
            registerOperation.TargetSE = target
            if operation.Catalog:
                registerOperation.Catalog = operation.Catalog

            targetSE = StorageElement(target, vo=self.vo)

            for ftsFile in ftsFileList:
                opFile = rmsFile()
                opFile.LFN = ftsFile.lfn
                opFile.Checksum = ftsFile.checksum
                # TODO: are we really ever going to change type... ?
                opFile.ChecksumType = "ADLER32"
                opFile.Size = ftsFile.size
                res = returnSingleResult(targetSE.getURL(ftsFile.lfn, protocol=registrationProtocols))

                # This should never happen !
                if not res["OK"]:
                    log.error("Could not get url", res["Message"])
                    continue
                opFile.PFN = res["Value"]
                registerOperation.addFile(opFile)

            request.insertBefore(registerOperation, operation)

        return self.reqClient.putRequest(request, useFailoverProxy=False, retryMainService=3)


class FTS3StagingOperation(FTS3Operation):
    """Class to be used for a Staging operation"""

    def prepareNewJobs(self, maxFilesPerJob=100, maxAttemptsPerFile=10):

        log = gLogger.getSubLogger("_prepareNewJobs")

        filesToSubmit = self._getFilesToSubmit(maxAttemptsPerFile=maxAttemptsPerFile)
        log.debug("%s ftsFiles to submit" % len(filesToSubmit))

        newJobs = []

        # {targetSE : [FTS3Files] }
        filesGroupedByTarget = FTS3Utilities.groupFilesByTarget(filesToSubmit)

        for targetSE, ftsFiles in filesGroupedByTarget.items():

            res = self._checkSEAccess(targetSE, "ReadAccess", vo=self.vo)
            if not res["OK"]:
                log.error(res)
                continue

            for ftsFilesChunk in breakListIntoChunks(ftsFiles, maxFilesPerJob):

                newJob = self._createNewJob("Staging", ftsFilesChunk, targetSE, sourceSE=targetSE)
                newJobs.append(newJob)

        return S_OK(newJobs)

    def _callback(self):
        """ " After a Staging operation, we have to update the matching Request in the
        RMS, and nothing more. If a callback is to be performed, it will be the next
        operation in the request, and put by the caller

        NOTE: we don't use ReqProxy when putting the request back to avoid operational hell
        """

        res = self._updateRmsOperationStatus()

        if not res["OK"]:
            return res

        request = res["Value"]["request"]

        return self.reqClient.putRequest(request, useFailoverProxy=False, retryMainService=3)



"""
  The InputData Optimizer Executor queries the file catalog for specified job input data and adds the
  relevant information to the job optimizer parameters to be used during the scheduling decision.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import pprint
import time

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.WorkloadManagementSystem.Executor.Base.OptimizerExecutor import OptimizerExecutor
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus


class InputData(OptimizerExecutor):
    """
    The specific Optimizer must provide the following methods:
      - initializeOptimizer() before each execution cycle
      - optimizeJob() - the main method called for each job
    """

    @classmethod
    def initializeOptimizer(cls):
        """Initialize specific parameters for InputData executor."""
        cls.failedMinorStatus = cls.ex_getOption("/FailedJobStatus", "Input Data Not Available")
        # this will ignore failover SE files
        cls.checkFileMetadata = cls.ex_getOption("CheckFileMetadata", True)
        # flag to require Input Data lookup with a user proxy
        cls.checkWithUserProxy = cls.ex_getOption("CheckWithUserProxy", False)

        cls.__dataManDict = {}
        cls.__fcDict = {}
        cls.__SEToSiteMap = {}
        cls.__lastCacheUpdate = 0
        cls.__cacheLifeTime = 600

        # Note: this is a default, that right now is generically the default for user jobs, at least for main DIRAC users
        # (since this now doesn't run for production jobs)
        # But this should probably be replaced by what the job actually request.
        # The problem is that the InputDataPolicy is not easy to get (a JDL parameter).
        # This may be used but clear how now
        # cls.__connectionLevel = 'PROTOCOL'

        return S_OK()

    def __getDataManager(self, vo):
        if vo in self.__dataManDict:
            return self.__dataManDict[vo]

        try:
            self.__dataManDict[vo] = DataManager(vo=vo)
        except Exception:
            msg = "Failed to create DataManager"
            self.log.exception(msg)
            return None
        return self.__dataManDict[vo]

    def __getFileCatalog(self, vo):
        if vo in self.__fcDict:
            return self.__fcDict[vo]

        try:
            self.__fcDict[vo] = FileCatalog(vo=vo)
        except Exception:
            msg = "Failed to create FileCatalog"
            self.log.exception(msg)
            return None
        return self.__fcDict[vo]

    def optimizeJob(self, jid, jobState):
        """This is the method that needs to be implemented by each and every Executor

        This optimizer will run if and only if it is needed:
          - it will run only if there are input files
          - for production jobs this can be skipped,
            since the logic is already applied by the transformation system, via the TaskManagerPlugins
        """
        # Is it a production job?
        result = jobState.getAttribute("JobType")
        if not result["OK"]:
            self.jobLog.error("Could not retrieve job type", result["Message"])
            return result
        jobType = result["Value"]
        if jobType in Operations().getValue("Transformations/DataProcessing", []):
            self.jobLog.info("Skipping optimizer, since this is a Production job")
            return self.setNextOptimizer()

        # Is there input data or not?
        result = jobState.getInputData()
        if not result["OK"]:
            self.jobLog.error("Cannot retrieve input data", result["Message"])
            return result
        inputData = result["Value"]

        result = self._getInputSandbox(jobState)
        if not result["OK"]:
            self.jobLog.error("Cannot retrieve input sandbox", result["Message"])
            return result
        inputSandbox = result["Value"]

        if not inputData and not inputSandbox:
            self.jobLog.notice("No input data nor LFN input sandboxes. Skipping.")
            return self.setNextOptimizer()

        # From now on we know that it is a user job with input data
        # and or with input sandbox

        # Check if we already executed this Optimizer and the input data is resolved
        result = self.retrieveOptimizerParam(self.ex_getProperty("optimizerName"))
        if result["OK"] and result["Value"]:
            self.jobLog.info("InputData optimizer ran already")
        else:
            self.jobLog.info("Processing input data")
            if self.checkWithUserProxy:
                result = jobState.getAttribute("Owner")
                if not result["OK"]:
                    self.jobLog.error("Could not retrieve job owner", result["Message"])
                    return result
                userName = result["Value"]
                result = jobState.getAttribute("OwnerGroup")
                if not result["OK"]:
                    self.jobLog.error("Could not retrieve job owner group", result["Message"])
                    return result
                userGroup = result["Value"]
                if inputSandbox:
                    result = self._resolveInputSandbox(  # pylint: disable=unexpected-keyword-arg
                        jobState, inputSandbox, proxyUserName=userName, proxyUserGroup=userGroup, executionLock=True
                    )
                    if not result["OK"]:
                        self.jobLog.error("Could not resolve input sandbox", result["Message"])
                        return result

                if inputData:
                    result = self._resolveInputData(  # pylint: disable=unexpected-keyword-arg
                        jobState, inputData, proxyUserName=userName, proxyUserGroup=userGroup, executionLock=True
                    )
            else:
                if inputSandbox:
                    result = self._resolveInputSandbox(jobState, inputSandbox)
                    if not result["OK"]:
                        self.jobLog.error("Could not resolve input sandbox", result["Message"])
                        return result

                if inputData:
                    result = self._resolveInputData(jobState, inputData)
            if not result["OK"]:
                self.jobLog.warn(result["Message"])
                return result

        return self.setNextOptimizer()

    #############################################################################

    def _getInputSandbox(self, jobState):
        """Return the LFN input sandbox if any

        :param JobState jobState: the JobState object
        :returns: S_OK/S_ERROR structure with (input sandbox lfn list)
        """
        inputSandbox = []
        # Check if the InputSandbox contains LFNs, and in that case treat them as input data
        result = jobState.getManifest()
        if not result["OK"]:
            return result
        manifest = result["Value"]
        # isb below will look something horrible like "['/an/lfn/1.txt', 'another/one.boo' ]"
        isb = manifest.getOption("InputSandbox")
        if not isb:
            return S_OK(inputSandbox)
        isbList = [li.replace("[", "").replace("]", "").replace("'", "") for li in isb.replace(" ", "").split(",")]
        for li in isbList:
            if li.startswith("LFN:"):
                inputSandbox.append(li.replace("LFN:", ""))
        return S_OK(inputSandbox)

    @executeWithUserProxy
    def _resolveInputData(self, jobState, inputData):
        """This method checks the file catalog for replica information.

        :param JobState jobState: JobState object
        :param list inputData: list of LFNs

        :returns: S_OK/S_ERROR structure with resolved input data info
        """
        lfns = inputData

        result = jobState.getManifest()
        if not result["OK"]:
            self.jobLog.error("Failed to get job manifest", result["Message"])
            return result
        manifest = result["Value"]
        vo = manifest.getOption("VirtualOrganization")
        startTime = time.time()
        dm = self.__getDataManager(vo)
        if dm is None:
            return S_ERROR("Failed to instantiate DataManager for vo %s" % vo)
        else:
            # This will return already active replicas, excluding banned SEs, and
            # removing tape replicas if there are disk replicas
            result = dm.getReplicasForJobs(lfns)
        self.jobLog.verbose("Catalog replicas lookup time", "%.2f seconds " % (time.time() - startTime))
        if not result["OK"]:
            self.log.warn(result["Message"])
            return result

        replicaDict = result["Value"]

        self.jobLog.verbose("REPLICA DICT", replicaDict)

        result = self.__checkReplicas(replicaDict, vo)

        if not result["OK"]:
            self.jobLog.error("Failed to check replicas", result["Message"])
            return result
        okReplicas = result["Value"]

        result = self.__getSiteCandidates(okReplicas, vo)
        if not result["OK"]:
            self.jobLog.error("Failed to check SiteCandidates", result["Message"])
            return result
        siteCandidates = result["Value"]

        if self.ex_getOption("CheckFileMetadata", True):
            result = jobState.getManifest()
            if not result["OK"]:
                return result
            manifest = result["Value"]
            vo = manifest.getOption("VirtualOrganization")
            fc = self.__getFileCatalog(vo)
            if fc is None:
                return S_ERROR("Failed to instantiate FileCatalog for vo %s" % vo)
            else:
                guidDict = fc.getFileMetadata(lfns)
            self.jobLog.info("Catalog Metadata Lookup Time", "%.2f seconds " % (time.time() - startTime))

            if not guidDict["OK"]:
                self.log.warn(guidDict["Message"])
                return guidDict

            failed = guidDict["Value"]["Failed"]
            if failed:
                self.log.warn("Failed to establish some GUIDs")
                self.log.warn(failed)

            for lfn in replicaDict["Successful"]:
                replicas = replicaDict["Successful"][lfn]
                guidDict["Value"]["Successful"][lfn].update(replicas)

        resolvedData = {}
        resolvedData["Value"] = guidDict
        resolvedData["SiteCandidates"] = siteCandidates
        self.jobLog.debug("Storing:\n%s" % pprint.pformat(resolvedData))
        result = self.storeOptimizerParam(self.ex_getProperty("optimizerName"), resolvedData)
        if not result["OK"]:
            self.log.warn(result["Message"])
            return result
        return S_OK(resolvedData)

    #############################################################################
    def __checkReplicas(self, replicaDict, vo):
        """Check that all input lfns have valid replicas and can all be found at least in one single site.

        :returns: S_ERROR/S_OK(dict of ok replicas)
        """
        badLFNs = []

        if "Successful" not in replicaDict:
            return S_ERROR("No replica Info available")

        okReplicas = replicaDict["Successful"]
        for lfn in okReplicas:
            if not okReplicas[lfn]:
                badLFNs.append("LFN:%s -> No replicas available" % (lfn))

        if "Failed" in replicaDict:
            errorReplicas = replicaDict["Failed"]
            for lfn in errorReplicas:
                badLFNs.append("LFN:%s -> %s" % (lfn, errorReplicas[lfn]))

        if badLFNs:
            errorMsg = "\n".join(badLFNs)
            self.jobLog.info("Found a number of problematic LFN(s)", "%d\n: %s" % (len(badLFNs), errorMsg))
            result = self.storeOptimizerParam(self.ex_getProperty("optimizerName"), errorMsg)
            if not result["OK"]:
                self.log.error("Failed to set job parameter", result["Message"])
            return S_ERROR("Input data not available")

        return S_OK(okReplicas)

    #############################################################################
    def __getSitesForSE(self, seName):
        """Returns a list of sites having the given SE as a local one.
        Uses the local cache of the site-se information
        """

        # Empty the cache if too old
        now = time.time()
        if (now - self.__lastCacheUpdate) > self.__cacheLifeTime:
            self.log.verbose("Resetting the SE to site mapping cache")
            self.__SEToSiteMap = {}
            self.__lastCacheUpdate = now

        if seName not in self.__SEToSiteMap:
            result = DMSHelpers().getSitesForSE(seName)
            if not result["OK"]:
                self.jobLog.error("Failed to get site for SE", result["Message"])
                return result
            self.__SEToSiteMap[seName] = list(result["Value"])
        return S_OK(self.__SEToSiteMap[seName])

    #############################################################################
    def __getSiteCandidates(self, okReplicas, vo):
        """This method returns a list of possible site candidates based on the job input data requirement.

        For each site candidate, the number of files on disk and tape is resolved.
        """

        lfnSEs = {}
        for lfn in okReplicas:
            replicas = okReplicas[lfn]
            siteSet = set()
            for seName in replicas:
                result = self.__getSitesForSE(seName)
                if not result["OK"]:
                    self.jobLog.warn("Could not get sites for SE", "%s: %s" % (seName, result["Message"]))
                    return result
                siteSet.update(result["Value"])
            lfnSEs[lfn] = siteSet

        if not lfnSEs:
            return S_ERROR(JobMinorStatus.NO_CANDIDATE_SITE_FOUND)

        # This makes an intersection of all sets in the dictionary and returns a set with it
        siteCandidates = set.intersection(*[lfnSEs[lfn] for lfn in lfnSEs])

        if not siteCandidates:
            return S_ERROR(JobMinorStatus.NO_CANDIDATE_SITE_FOUND)

        # In addition, check number of files on tape and disk for each site
        # for optimizations during scheduling
        sitesData = {}
        for siteName in siteCandidates:
            sitesData[siteName] = {"disk": set(), "tape": set()}

        # Loop time!
        seDict = {}
        for lfn in okReplicas:
            replicas = okReplicas[lfn]
            # Check each SE in the replicas
            for seName in replicas:
                # If not already "loaded" the add it to the dict
                if seName not in seDict:
                    result = self.__getSitesForSE(seName)
                    if not result["OK"]:
                        self.jobLog.warn("Could not get sites for SE", "%s: %s" % (seName, result["Message"]))
                        continue
                    siteList = result["Value"]
                    seObj = StorageElement(seName, vo=vo)
                    result = seObj.getStatus()
                    if not result["OK"]:
                        self.jobLog.error("Failed to get SE status", result["Message"])
                        return result
                    seDict[seName] = {"Sites": siteList, "Status": result["Value"]}
                # Get SE info from the dict
                seData = seDict[seName]
                siteList = seData["Sites"]
                seStatus = seData["Status"]
                for siteName in siteList:
                    # If not a candidate site then skip it
                    if siteName not in siteCandidates:
                        continue
                    # Add the LFNs to the disk/tape lists
                    diskLFNs = sitesData[siteName]["disk"]
                    tapeLFNs = sitesData[siteName]["tape"]
                    if seStatus["DiskSE"]:
                        # Sets contain only unique elements, no need to check if it's there
                        diskLFNs.add(lfn)
                        if lfn in tapeLFNs:
                            tapeLFNs.remove(lfn)
                    if seStatus["TapeSE"]:
                        if lfn not in diskLFNs:
                            tapeLFNs.add(lfn)

        for siteName in sitesData:
            sitesData[siteName]["disk"] = len(sitesData[siteName]["disk"])
            sitesData[siteName]["tape"] = len(sitesData[siteName]["tape"])
        return S_OK(sitesData)

    @executeWithUserProxy
    def _resolveInputSandbox(self, jobState, inputSandbox):
        """This method checks the file catalog for replica information.

        :param JobState jobState: JobState object
        :param list inputSandbox: list of LFNs for the input sandbox

        :returns: S_OK/S_ERROR structure with resolved input data info
        """

        result = jobState.getManifest()
        if not result["OK"]:
            self.jobLog.error("Failed to get job manifest", result["Message"])
            return result
        manifest = result["Value"]
        vo = manifest.getOption("VirtualOrganization")
        startTime = time.time()
        dm = self.__getDataManager(vo)
        if dm is None:
            return S_ERROR("Failed to instantiate DataManager for vo %s" % vo)

        # This will return already active replicas, excluding banned SEs, and
        # removing tape replicas if there are disk replicas

        result = dm.getReplicasForJobs(inputSandbox)
        self.jobLog.verbose("Catalog replicas lookup time", "%.2f seconds " % (time.time() - startTime))
        if not result["OK"]:
            self.log.warn(result["Message"])
            return result

        isDict = result["Value"]

        self.jobLog.verbose("REPLICA DICT", isDict)

        result = self.__checkReplicas(isDict, vo)

        if not result["OK"]:
            self.jobLog.error("Failed to check replicas", result["Message"])

        return result
